﻿// Network Table values:
//
//      shooter_desiredSpeed            shooter speed 0.00 to 100.00
//      shooter_desiredSpeed_e          what user wants the shooter speed to be: 0.00 to 100.00
//
//      vision_Start                    robot tells dashboard to start/stop vision procesing: true/false
//      vision_Counter                  gets incremented (by the dashboard) each time an image is processed, starting at 1
//      vision_HorizontalOffset         how far off center the entire target is from center of image, in pixels (used in: 2017)
//      vision_DistanceBetween          how far apart are two targets, in pixels (used in: 2017)
//      vision_Result                   0 - no targets at all; 1 - valid target; 2 - not a valid target (used in: 2017)


using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ThunderDashboard
{
    public partial class Form2017 : FormBase
    {
        public Form2017()
        {
            InitializeComponent();

            Text = "Thunder Shogun"; // Robot name goes here (i.e. tab's text)
            DebugTitle = "Debug"; // title for the debug tab

            // load controller images
            ControllerAux = LoadImage("2017", "ControlsAux.png");
            ControllerDriver = LoadImage("2017", "ControlsDriver.png");

            // optional tabs
            TabLightsPsoc42xx = true;   // show the light tab that talks to the PSOC 42xx via a COM port?
            TabChart = true;            // show the chart tab?



            // *******************************************************************************
            // any custom control settings for this year's robot

            // do not allow user to update power (sent to robot) until the click checkbox
            numericUpDownPower.Enabled = false;

            //
            // *******************************************************************************



            // *******************************************************************************
            // NOTE: if a control/feature is not supported (i.e. control does not exist), then
            //       set the control to null (e.g. CameraFeed = null;)
            // *******************************************************************************

            // camera
            CameraPort = 1181; // port the camera feed is on (as of 2018 it is MPEJ on 1181, previous roboRIO years it was 1180)
            CameraFlip = true; // camera is mounted upside down, so flip image before showing it?
            CameraFeed = cameraControl; // control that gets the camera feed
            CameraVisionDelegate = VisionProcessing; // method to call if there is any vision processing that the dashboard will do
            CameraFeedDirection = labelCameraSide; // indicates if camera feed is from the front or rear of robot (only needed is two cameras on robot)

            // drive
            RobotDirection = labelReverseDrive; // will so REVERSE when driver controls have been swapped so rear of robot is the 'front'

            // auto mode
            AutoModeComboBox = comboBoxAutoMode; // gets list of auto modes for user to select from
            AutoModeDescriptionsTextBox = autoModeDescriptionsBox; // gets the description of the auto mode (not used in 2017)
            AutoModeFromRobot = labelAutoModeValue; // auto mode the robot currently has selected

            // connection/status
            PanelConnected = panelConnected; // state of connection with network table
            StatusTextBox = statusTextBox; // any status messages about connection and images
        }


        //
        //      vision_Result                   0 - no targets at all; 1 - valid target; 2 - not a valid target (used in: 2017)
        private bool VisionProcessing(ThunderVision.TARGET_INFO[] targetInfo, int targets, int imageWidth, int imageHeight)
        {
            int distanceBetween = 0;
            int leftCenter = 0;
            int rightCenter = 0;
            int horizontalOffset = 0;
            int result = 0;
            List<ThunderVision.TARGET_INFO> validTargets = new List<ThunderVision.TARGET_INFO>();

            // if camera is looking too far down, we might get a flare from the gear
            // (reflection from the yellow can look green). but when that is the case,
            // the two field targets will be higher in the image.
            // if we have a gear flare, we should only ever have one flare; this would
            // be because the camera is not centered within the gap between the two gear
            // teeth and the flare would be off the tooth closest to the camera.
            // however, we need to be careful for the case of the camera looking too
            // far up and the two field targets are touching the bottom of the image.
            //
            // start by seeing how many are touching the bottom
            int onBottom = 0;
            for (int i = 0; i < targets; i++)
            {
                if (targetInfo[i].bottom >= (imageHeight - 10))
                {
                    onBottom++;
                }
            }
            // now process the targets
            for (int i = 0; i < targets; i++)
            {
                // if only one on the bottom and this is it, then skip it
                if ((onBottom == 1) && (targetInfo[i].bottom >= (imageHeight - 10)))
                {
                    targetInfo[i].valid = false;        // make it as being invalid
                }
                else if (targetInfo[i].top < 10)
                {
                    targetInfo[i].valid = false;        // make it as being invalid
                }
                else
                {
                    validTargets.Add(targetInfo[i]);    // keep it because it is within the image
                }
            }

            // see if the targets are valid
            targets = validTargets.Count;
            if (targets < 1)
            {
                result = 0;                     // no targets found
            }
            else if (targets != 2)
            {
                result = 2;                     // we need two targets
            }
            else if (validTargets[0].row != validTargets[1].row)
            {
                result = 2;                     // they must be next to each other
            }
            else
            {
                // we have valid targets
                result = 1;

                // calc based on which one is to the real left
                if (validTargets[0].left <= validTargets[1].left)
                {
                    // first target is on the left
                    distanceBetween = validTargets[1].left - validTargets[0].right;
                    leftCenter = validTargets[0].left + (validTargets[0].maxWidth / 2);
                    rightCenter = validTargets[1].left + (validTargets[1].maxWidth / 2);
                }
                else
                {
                    // first target is on the right
                    distanceBetween = validTargets[0].left - validTargets[1].right;
                    leftCenter = validTargets[1].left + (validTargets[1].maxWidth / 2);
                    rightCenter = validTargets[0].left + (validTargets[0].maxWidth / 2);
                }
                horizontalOffset = ((rightCenter - leftCenter) / 2) + leftCenter;
                horizontalOffset = (horizontalOffset - (imageWidth / 2));
            }

            DashboardNetworkTable.PutNumber("vision_DistanceBetween", (double)distanceBetween);
            DashboardNetworkTable.PutNumber("vision_HorizontalOffset", (double)horizontalOffset);
            DashboardNetworkTable.PutNumber("vision_Result", (double)result);
            DashboardNetworkTable.PutNumber("vision_Counter", (double)_visionCounter++);
            return (result == 1);
        }

        // called with Reset button is pressed
        public override void buttonReset_Click(object sender, EventArgs e)
        {
            // just let base class to it all
            base.buttonReset_Click(sender, e);
        }

        // called when the network table is updated
        //
        // returns true if the form handled the key
        public override bool NetworkTableUpdated(string key, string value)
        {
            switch (key)
            {
                default:
                    return (base.NetworkTableUpdated(key, value));

                case "shooter_desiredSpeed":
                    meterControlPower.Value = value;
                    break;

                case "shooter_desiredSpeed_e":
                    try
                    {
                        double speed = Convert.ToDouble(value);
                        if (speed != Convert.ToDouble(numericUpDownPower.Value))
                        {
                            _speed = speed;
                            numericUpDownPower.Value = Convert.ToDecimal(speed);
                        }
                    }
                    catch
                    {
                    }
                    break;
            }

            return (false);
        }

        // called when 'Enable Speed Changer' check is changed
        private void checkBoxPower_CheckedChanged(object sender, EventArgs e)
        {
            numericUpDownPower.Enabled = checkBoxPower.Checked;
        }

        // called when 'Speed Changer' number is changed
        private void numericUpDownPower_ValueChanged(object sender, EventArgs e)
        {
            try
            {
                double speed = Convert.ToDouble(numericUpDownPower.Value);
                if (speed != _speed)
                {
                    _speed = speed;
                    DashboardNetworkTable.PutNumber("shooter_desiredSpeed_e", _speed);
                }
            }
            catch
            {
            }
        }
        private double _speed = 0;
    }
}
