﻿// Network Table values:
//
//      thunderdashboard_armpos             position of arm 0=all the way back, 100=all the way front
//      thunderdashboard_intake             1=intake has a cube, 0=no cube
//      thunderdashboard_auto_start_delay   number seconds, 0 to 15, to delay start of auto mode; force default to 0


using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ThunderDashboard
{
    public partial class Form2018 : FormBase
    {
        public Form2018()
        {
            InitializeComponent();

            Text = "Super Cube Finder II Turbo"; // Robot name goes here (i.e. tab's text)
            DebugTitle = "Kate and Josh's Savior"; // title for the debug tab

            // load controller images
            ControllerAux = LoadImage("2018", "ControlsAux.png");
            ControllerDriver = LoadImage("2018", "ControlsDriver.png");

            // optional tabs
            TabLightsPsoc42xx = true;   // show the light tab that talks to the PSOC 42xx via a COM port?
            TabChart = true;            // show the chart tab?



            // *******************************************************************************
            // any custom control settings for this year's robot
            autoDelayWritten = false;       // we did not intialize the dashboard with a value yet
            textBoxAutoDelay.Text = "0";    // default to 0 seconds
            //
            // *******************************************************************************



            // *******************************************************************************
            // NOTE: if a control/feature is not supported (i.e. control does not exist), then
            //       set the control to null (e.g. CameraFeed = null;)
            // *******************************************************************************

            // camera
            CameraPort = 1181; // port the camera feed is on (as of 2018 it is MPEJ on 1181, previous roboRIO years it was 1180)
            CameraFlip = false; // camera is mounted upside down, so flip image before showing it?
            CameraFeed = cameraControl; // control that gets the camera feed
            CameraVisionDelegate = null; // method to call if there is any vision processing that the dashboard will do
            CameraFeedDirection = null; // indicates if camera feed is from the front or rear of robot (only needed if two cameras on robot)

            // drive
            RobotDirection = labelReverseDrive; // will so REVERSE when driver controls have been swapped so rear of robot is the 'front'

            // auto mode
            AutoModeComboBox = comboBoxAutoMode; // gets list of auto modes for user to select from
            AutoModeDescriptionsTextBox = autoModeDescriptionsBox; // gets the description of the auto mode (not used in 2017)
            AutoModeFromRobot = labelAutoModeValue; // auto mode the robot currently has selected

            // connection/status
            PanelConnected = panelConnected; // state of connection with network table
            StatusTextBox = statusTextBox; // any status messages about connection and images
        }

        // called with Reset button is pressed
        public override void buttonReset_Click(object sender, EventArgs e)
        {
            // just let base class to it all
            base.buttonReset_Click(sender, e);
        }

        // called when the network table is updated
        //
        // returns true if the form handled the key
        public override bool NetworkTableUpdated(string key, string value)
        {
            switch (key)
            {
                default:
                    // if we just connected, then set values that the dashboard drives
                    if (key == DashboardNetworkTable.Connected)
                    {
                        if (!autoDelayWritten)
                        {
                            autoDelayWritten = true;
                            textBoxAutoDelay_TextChanged(null, null);// simulate user changing setting
                        }
                    }
                    return (base.NetworkTableUpdated(key, value));

                case "thunderdashboard_armpos":
                    meterControlPower.Value = value;
                    break;

                case "thunderdashboard_intake":
                    if (value == "1")
                    {
                        panelIntake.BackColor = Color.Yellow;
                    }
                    else
                    {
                        panelIntake.BackColor = SystemColors.Control;
                    }
                    break;

                case "thunderdashboard_auto_start_delay":
                    // if value is 0, but we are showing a blank, then do not
                    // change the control; otherwise the user won't be able
                    // to delete the enter and type a number
                    if ((value == "0") && (textBoxAutoDelay.Text == ""))
                    {
                        break;
                    }
                    textBoxAutoDelay.Text = value;
                    break;
            }

            return (false);
        }

        // called when the Auto Delay value is changed
        private void textBoxAutoDelay_TextChanged(object sender, EventArgs e)
        {
            string value = textBoxAutoDelay.Text;
            string validated = "";
            int lengthToCheck;

            // valid the length. while we set the control's property to prevent
            // the user from putting in too many characters, it could be too
            // long if network tables sent in too long of a value
            lengthToCheck = value.Length;
            if (value.Length > textBoxAutoDelay.MaxLength)
            {
                lengthToCheck = textBoxAutoDelay.MaxLength;
            }

            // make sure each character is a digit. if not, then put back just the digits
            for (int i = 0; i < lengthToCheck; i++)
            {
                if ((value[i] >= 0x30) && (value[i] <= 0x39))
                {
                    validated += value[i];
                }
            }
            if (validated != value)
            {
                textBoxAutoDelay.Text = validated;
                textBoxAutoDelay.Select(textBoxAutoDelay.Text.Length, 0);
            }

            // see what the value is and give it to the dashboard
            int delay;
            try
            {
                if (textBoxAutoDelay.Text == "")
                {
                    delay = 0;
                }
                else
                {
                    delay = Convert.ToInt32(textBoxAutoDelay.Text);
                }
            }
            catch
            {
                delay = 0;
            }

            // if there is a delay, let's 'highlight' the auto mode selection so they see it is different
            if (delay == 0)
            {
                comboBoxAutoMode.FlatStyle = FlatStyle.Standard;
                comboBoxAutoMode.BackColor = SystemColors.Window;
            }
            else
            {
                comboBoxAutoMode.FlatStyle = FlatStyle.Flat;
                comboBoxAutoMode.BackColor = Color.Yellow;
            }

            // update robot
            DashboardNetworkTable.PutNumber("thunderdashboard_auto_start_delay", Convert.ToDouble(delay));
        }

        // member variables
        private bool autoDelayWritten = false;  // did we give the dashboard the value yet?
    }
}
