﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ThunderDashboard
{
    public class AutoMode
    {
        // set controls to be used and initialize them
        public void Initialize(ComboBox comboBoxAutoSelection, TextBox textBoxAutoDescription, string autoModeKey = "Auto_Mode", string keyPrefix = "thunderdashboard_auto_")
        {
            _comboBoxAutoSelection = comboBoxAutoSelection;
            if (_comboBoxAutoSelection != null)
            {
                _comboBoxAutoSelection.SelectedIndexChanged += new System.EventHandler(comboBoxAutoSelection_SelectedIndexChanged);
            }

            _textBoxAutoDescription = textBoxAutoDescription;

            _autoModeKey = autoModeKey;

            _keyPrefix = keyPrefix;

            SetList(Properties.Settings.Default.AutoList);
        }

        // call when network tables is connected, so we get auto mode from robot
        public void Connected()
        {
            SetList(Properties.Settings.Default.AutoList);
        }

        // call when a auto key (i.e. _keyPrefix) has been changed via the network tables
        public void UpdateValue(string key, string value)
        {
            // make sure it is a key we care about
            if (!key.StartsWith(_keyPrefix))
            {
                return;
            }

            // first see if we got a list
            if (key == (_keyPrefix + "list"))
            {
                SetList(value);
                return;
            }

            // not a list, so assume it is a description
            SetDescription(key, value);
        }




        // will take the list of modes from the robot and list easch one
        private void SetList(string list)
        {
            string[] modes;
            //string selectedItem;
            int selectedMode;

            // make sure the combobox was set
            if (_comboBoxAutoSelection == null)
            {
                return;
            }

            // if no list, then make an empty list
            if (list == null)
            {
                modes = new string[] { };
            }
            else
            {
                modes = list.Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries);
            }

            // get the current selection, so we can put it back
            if (_comboBoxAutoSelection.SelectedIndex >= 0)
            {
                selectedMode = ModeFromString(_comboBoxAutoSelection.SelectedItem.ToString());
            }
            else
            {
                selectedMode = Properties.Settings.Default.AutoMode;
            }

            _comboBoxAutoSelection.SuspendLayout();
            _comboBoxAutoSelection.Items.Clear();
            if (modes.Length <= 0)
            {
                // nothing yet, so just go with "0 - do nothing"
                _comboBoxAutoSelection.Items.Add("0");
                Properties.Settings.Default.AutoMode = 0;
            }
            else
            {
                for (int i = 0; i < modes.Length; i++)
                {
                    _comboBoxAutoSelection.Items.Add(ModeToString(modes[i]));
                }
            }
            _comboBoxAutoSelection.ResumeLayout();

            // don't re-save what we just read from the file
            if (list != Properties.Settings.Default.AutoList)
            {
                try
                {
                    Properties.Settings.Default.AutoList = list;
                    Properties.Settings.Default.Save();
                }
                catch
                {
                }
            }

            // put back the current selection. if not there, then pick the first one
            _comboBoxAutoSelection.SelectedItem = ModeToString(selectedMode.ToString());
            if ((_comboBoxAutoSelection.SelectedIndex < 0) && (_comboBoxAutoSelection.Items.Count > 0))
            {
                _comboBoxAutoSelection.SelectedIndex = 0;
            }
        }
        private string ModeToString(string number)
        {
            string desc = GetAutoText(ModeFromString(number));
            return (number + ": " + desc);
        }

        // will take a auto description and update it 
        private void SetDescription(string key, string value)
        {
            StringCollection keyValues = Properties.Settings.Default.AutoDesc;
            bool changed = false;

            // if nothing created yet, do that now
            if (keyValues == null)
            {
                // first item for the collection
                keyValues = new StringCollection();
                keyValues.Add(key);
                keyValues.Add(value);
                changed = true;
            }
            else
            {
                // we have a collection, so find out string
                int idx = keyValues.IndexOf(key);
                if (idx < 0)
                {
                    // not in there yet, so add it
                    keyValues.Add(key);
                    keyValues.Add(value);
                    changed = true;
                }
                else if (idx >= keyValues.Count)
                {
                    keyValues.Add(value);
                    changed = true;
                }
                else
                {
                    // replace the string
                    if (keyValues[idx + 1] != value)
                    {
                        keyValues[idx + 1] = value;
                        changed = true;
                    }
                }
            }

            // only update the settings if something changed
            if (changed)
            {
                try
                {
                    Properties.Settings.Default.AutoDesc = keyValues;
                    Properties.Settings.Default.Save();
                    comboBoxAutoSelection_SelectedIndexChanged(null, null);
                }
                catch
                {
                }
            }
        }

        // get an auto description
        private string GetAutoText(int autoMode)
        {
            StringCollection keyValues = Properties.Settings.Default.AutoDesc;

            // if nothing set yet, just return
            if (keyValues == null)
            {
                return ("");
            }

            // find the item we need the text for
            int idx = keyValues.IndexOf(_keyPrefix + autoMode.ToString());

            // if item not found (or no value after it), then just return
            if ((idx < 0) || (idx >= keyValues.Count))
            {
                // not in there yet, so add it
                return ("");
            }
            return (keyValues[idx + 1]);
        }

        // called when the auto mode list selection changes
        private int ModeFromString(string mode)
        {
            try
            {
                string[] strs = mode.Split(new string[] { ": " }, StringSplitOptions.None);

                return (Convert.ToInt32(strs[0]));
            }
            catch
            {
            }
            return (0);
        }
        private void comboBoxAutoSelection_SelectedIndexChanged(object sender, EventArgs e)
        {
            int autoMode = 0;

            try
            {
                // save sure something was selected
                if (_comboBoxAutoSelection.SelectedIndex >= 0)
                {
                    autoMode = ModeFromString(_comboBoxAutoSelection.SelectedItem.ToString());
                }
            }
            catch
            {
            }
            if (_textBoxAutoDescription != null)
            {
                _textBoxAutoDescription.Text = GetAutoText(autoMode);
            }

            if (_autoModeKey != null)
            {
                DashboardNetworkTable.PutNumber(_autoModeKey, (double)autoMode);
            }

            // save the item
            try
            {
                Properties.Settings.Default.AutoMode = autoMode;
                Properties.Settings.Default.Save();
            }
            catch
            {
            }
        }

        private ComboBox _comboBoxAutoSelection = null;
        private TextBox _textBoxAutoDescription = null;
        private string _autoModeKey = null;
        private string _keyPrefix = "";
    }
}
