﻿// handle IP cameras

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net.Sockets;
using System.Threading;
using System.Net;
using System.Collections;
using System.IO;
using System.Drawing;
using System.ComponentModel;

namespace ThunderDashboard
{
    partial class CameraControl : PictureBox
    {
        // runs the camera (for background worker) until told to exit camera
        // Returns: true if we should wait and retry again
        private bool RunCamera()
        {
            bool ret = false;

            // do our one-time initialization of the table for jpeg images
            if (huffman_table == null)
            {
                huffman_table = new byte[huffman_table_int.Length];
                for (int i = 0; i < huffman_table.Length; ++i)
                {
                    huffman_table[i] = (byte)huffman_table_int[i];
                }
            }

            client = null;					// make sure client is null
            while (!_shutdownCamera && !_resetCamera)		// needed incase we are stuck in outer try/catch
            {
                // Connect to remote camera.
                try
                {
                    // Establish the remote endpoint for the socket.
                    //Thread.Sleep(5000);
                    IPHostEntry ipHostInfo = Dns.GetHostEntry(_cameraIp);

                    foreach (IPAddress ipAddress in ipHostInfo.AddressList)
                    {
                        // We get back two addresses: IPv6 and IPv4. Only the IPv4 works for the
                        // socket connection
                        if (ipAddress.AddressFamily == AddressFamily.InterNetwork)
                        {
                            IPEndPoint remoteEP = new IPEndPoint(ipAddress, _port);

                            // Create a TCP/IP socket.
                            client = new Socket(AddressFamily.InterNetwork,
                                SocketType.Stream, ProtocolType.Tcp);
                            if (client == null)
                            {
                                //UpdateStatusDescription("socket fail");
                                break;
                            }

                            // Connect to the remote endpoint.
                            //UpdateStatusDescription("try connect");
                            connectDone.Reset();
                            client.BeginConnect(remoteEP, new AsyncCallback(ConnectCallback), client);
                            connectDone.WaitOne();
                            if (connectError)
                            {
                                //UpdateStatusDescription("connect fail");
                                client.Close();
                                client = null;

                                // break out and start all over looking for the robot
                                _resetCamera = true;
                                break;
                            }

                            // configure the camera
                            if (_resolution != Resolution.ResolutionDoNotSet)
                            {
                                int res = 0;
                                switch (_resolution) // frame size: 0=640x480, 1=320x240, 2=160x120
                                {
                                    case Resolution.Resolution640x480: res = 0; break;
                                    case Resolution.Resolution320x240: res = 1; break;
                                    case Resolution.Resolution160x120: res = 2; break;
                                }
                                // Send data to the roboRIO
                                // (frames per second, hardware compression, frame size)
                                sendDone.Reset();
                                if (!Send(client, new int[] { _fps, -1, res }))  
                                {
                                    //UpdateStatusDescription("send fail");
                                    //client.Shutdown(SocketShutdown.Both);
                                    //client.Close();
                                    client = null;
                                    break;
                                }
                                sendDone.WaitOne();
                                if (sendError)
                                {
                                    //UpdateStatusDescription("send done fail");
                                    client.Shutdown(SocketShutdown.Both);
                                    client.Close();
                                    client = null;
                                    break;
                                }
                            }

                            // Receive the response from the remote device.
                            while (!_shutdownCamera && !_resetCamera)
                            {
                                try
                                {
                                    // Look for the "magic" number that marks the beginning of an image
                                    if (!Receive(client, 4))
                                    {
                                        //UpdateStatusDescription("rec fail");
                                        break;
                                    }
                                    receiveDone.WaitOne();
                                    if (response == null)
                                    {
                                        //UpdateStatusDescription("rec 1 fail");
                                        break;  // lost connection
                                    }

                                    bool magicNumbersNotFound = true;
                                    bool connectionLost = false;

                                    while (magicNumbersNotFound && !_shutdownCamera && !_resetCamera && !connectionLost)
                                    {
                                        if (ByteArrayCompare(magicNumbers, response))
                                        {
                                            magicNumbersNotFound = false;

                                            // Found start of image

                                            // Read size of image
                                            if (!Receive(client, 4))
                                            {
                                                //UpdateStatusDescription("rec a fail");
                                                connectionLost = true;
                                                break;  // lost connection
                                            }
                                            receiveDone.WaitOne();
                                            if (response == null)
                                            {
                                                //UpdateStatusDescription("rec 2 fail");
                                                connectionLost = true;
                                                break;  // lost connection
                                            }
                                            int imageSize = intFromNetworkBytes(response);

                                            // Read whole image
                                            if (!Receive(client, imageSize))
                                            {
                                                //UpdateStatusDescription("rec b fail");
                                                connectionLost = true;
                                                break;  // lost connection
                                            }
                                            receiveDone.WaitOne();
                                            if (response == null)
                                            {
                                                //UpdateStatusDescription("rec 3 fail");
                                                connectionLost = true;
                                                break;  // lost connection
                                            }

                                            byte[] processedImage; // = new byte[imageSize + huffman_table.Length];
                                            processedImage = AddHuffmanTable(imageSize);
                                            if (processedImage != null)
                                            {
                                                UpdateStatusDescription("At " + DateTime.Now + " got image");
												//File.WriteAllBytes(@"C:\Users\Robotics\Documents\testimage.jpg", processedImage);
																												
												// make sure the image is updated in the UI thread
												Invoke(new UpdateImageDelegate(UpdateImage), processedImage);
                                            }
                                        }
                                        else // response doesn't match magic numbers
                                        {
                                            // Takes the last 3 bytes of the previous response, adds the next byte, and then compares the
                                            // new array to the magic numbers.
                                            prevResponse = response;
                                            byte[] tempResponse = new byte[4];
                                            if (!Receive(client, 1))
                                            {
                                                //UpdateStatusDescription("rec c fail");
                                                connectionLost = true;
                                                break;  // lost connection
                                            }
                                            receiveDone.WaitOne();
                                            if (response == null)
                                            {
                                                //UpdateStatusDescription("rec 4 fail");
                                                connectionLost = true;
                                                break;  // lost connection
                                            }
                                            Array.Copy(prevResponse, 1, tempResponse, 0, 3);
                                            tempResponse[3] = response[0];
                                            response = tempResponse;
                                        }
                                    }
                                }
                                catch (Exception e)
                                {
                                    Console.WriteLine(e.ToString());
                                    UpdateStatusDescription("At " + DateTime.Now + " Failed while finding magic numbers/writing image with error: " + e.ToString());

                                    // break out and start all over looking for the robot
                                    _resetCamera = true;
                                }
                                response = null;
                                //prevResponse = null;

                            } // while

                            if (client != null)
                            {
                                //UpdateStatusDescription("close it");
                                client.Shutdown(SocketShutdown.Both);
                                client.Close();
                                client.Dispose();	// MDF 3/11/2015: Seems unnecessary, but desperate
                                client = null;		// avoid closing twice
                            }
							break;	// get out of foreach loop; no need to waste time
                        } // if
                    } // foreach
                } // try
                catch (Exception e)
                {
                    Console.WriteLine(e.ToString());
                    UpdateStatusDescription("At " + DateTime.Now + " Failed while connecting to socket with error: " + e.ToString());

                    if (client != null)
                    {
                        client.Close();
                        client = null;        //MAB - added to avoid closing twice
                    }

                    // break out and start all over looking for the robot
                    _resetCamera = true;
                }
            }
            //UpdateStatusDescription("At " + DateTime.Now + " DEAD");

            return (ret);
        }

        private byte[] AddHuffmanTable(int imageSize)
        {
            byte[] processedImage = new byte[imageSize + huffman_table.Length];

            // Find if and where we need to insert the huffman table
            bool has_dht = false;
            int pos = 2;
			while (!has_dht && ((pos + 4) <= imageSize))
			{
				if ((response[pos + 1] & 0xff) == 0xc4)
				{
					has_dht = true;
				}
				else if ((response[pos + 1] & 0xff) == 0xda)
				{
					break;
				}

                // Skip to the next marker.
                int marker_size = ((response[pos + 2] & 0xff) << 8) + (response[pos + 3] & 0xff);
                pos += marker_size + 2;
            }

            if (!has_dht && (pos < imageSize)) // MDF 2/24/15: Added test for pos < imageSize, too
            {
                // We need to add huffman table at 'pos' - shift frame data
                // to make room
                try
                {
                    Array.Copy(response, 0, processedImage, 0, pos);
                    Array.Copy(response, pos, processedImage, pos + huffman_table.Length, imageSize - pos);
                    // Now copy in huffman table at 'pos'
                    Array.Copy(huffman_table, 0, processedImage, pos, huffman_table.Length);
                    // Upsize dataLen to include huffman table
                    imageSize += huffman_table.Length;
                }
                catch (Exception e)
                {
                    Console.WriteLine(e.ToString());
                    UpdateStatusDescription("At " + DateTime.Now + " Failed while inserting huffman table with error: " + e.ToString());

                }
            }
            else
            {
                // We will not be inserting the huffman table with this
                // frame
#if !HANDLE_TABLEFUL
                byte[] unProcessedImage = new byte[imageSize];
                Array.Copy (response, unProcessedImage, imageSize);
                return unProcessedImage;
#else
                return null;
#endif
            }

            return (processedImage);
        }

        private bool ByteArrayCompare(byte[] a1, byte[] a2)
        {
            return (StructuralComparisons.StructuralEqualityComparer.Equals(a1, a2));
        }

        private void ConnectCallback(IAsyncResult ar)
        {
            try
            {
                // Retrieve the socket from the state object.
                Socket client = (Socket)ar.AsyncState;

                // Complete the connection.
                client.EndConnect(ar);

                Console.WriteLine("Socket connected to {0}",
                    client.RemoteEndPoint.ToString());

                // Signal that the connection has been made.
                connectError = false;
                connectDone.Set();
            }
            catch (SocketException e)
            {
                Console.WriteLine(e.ToString());
                Console.WriteLine ("Err {0} port {1} ip {2}", e.ErrorCode, _port, _cameraIp);
                UpdateStatusDescription("At " + DateTime.Now + " Failed in ConnectionCallback with error: " + e.ToString());
                connectError = true;
                connectDone.Set();
            }
        }

        private bool Receive(Socket client, int bufferSize)
        {
            bool fSuccess = true;

            try
            {
                // Create the state object.
                StateObject state = new StateObject();
                state.buffer = new Byte[bufferSize];
                state.workSocket = client;

                state.bytesRequested = bufferSize;
                state.bytesReceived = 0;
                receiveDone.Reset();

                // Begin receiving the data from the remote device.
                client.BeginReceive(state.buffer, 0, bufferSize, 0,
                    new AsyncCallback(ReceiveCallback), state);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
                UpdateStatusDescription("At " + DateTime.Now + " Failed while receiving with error: " + e.ToString());

                fSuccess = false;
            }

            return fSuccess;
        }

        private void ReceiveCallback(IAsyncResult ar)
        {
            try
            {
                // Retrieve the state object and the client socket 
                // from the asynchronous state object.
                StateObject state = (StateObject)ar.AsyncState;
                Socket client = state.workSocket;

                // Read data from the remote device.
                int bytesRead = client.EndReceive(ar);

                if (bytesRead > 0)
                {
                    state.bytesReceived += bytesRead;

                    // There might be more data, so store the data received so far.
                    //state.sb.Append(Encoding.ASCII.GetString(state.buffer, 0, bytesRead));

                    // if not done, yet, get more
                    if (state.bytesReceived < state.bytesRequested)
                    {
                        // Get the rest of the data.
                        client.BeginReceive(state.buffer, state.bytesReceived, (state.bytesRequested - state.bytesReceived), 0,
                            new AsyncCallback(ReceiveCallback), state);
                        return;
                    }
                }

                // All the data has arrived; put it in response.
                //if (state.sb.Length > 1)
                if (state.bytesReceived >= state.bytesRequested)
                {
                    response = state.buffer;
                }
                else
                {
                    response = null;
                }
                // Signal that all bytes have been received.
                receiveDone.Set();
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
                UpdateStatusDescription("At " + DateTime.Now + " Failed in ReceiveCallback with error: " + e.ToString());
                response = null;
                receiveDone.Set();
            }
        }

        private void Send(Socket client, String data)
        {
            // Convert the string data to byte data using ASCII encoding.
            byte[] byteData = Encoding.ASCII.GetBytes(data);

            // Begin sending the data to the remote device.
            client.BeginSend(byteData, 0, byteData.Length, 0,
                new AsyncCallback(SendCallback), client);
        }

        private int intFromNetworkBytes(byte[] byteData)
        {
            int integer = 0;
            integer = integer | (((int)byteData[0] & 0xff) << 24);
            integer = integer | (((int)byteData[1] & 0xff) << 16);
            integer = integer | (((int)byteData[2] & 0xff) << 8);
            integer = integer | (((int)byteData[3] & 0xff));
            return integer;
        }

        private bool Send(Socket client, int[] data)
        {
            try
            {
                // Convert the string data to byte data using ASCII encoding.
                byte[] byteData = new byte[4 * data.Length];
                for (int i = 0; i < data.Length; i++)
                {
                    byteData[(i * 4) + 0] = (byte)((data[i] >> 24) & 255);
                    byteData[(i * 4) + 1] = (byte)((data[i] >> 16) & 255);
                    byteData[(i * 4) + 2] = (byte)((data[i] >> 8) & 255);
                    byteData[(i * 4) + 3] = (byte)(data[i] & 255);
                }

                // Begin sending the data to the remote device.
                client.BeginSend(byteData, 0, byteData.Length, 0,
                    new AsyncCallback(SendCallback), client);
                return (true);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
                UpdateStatusDescription("At " + DateTime.Now + " Failed while sending with error: " + e.ToString());

                return (false);
            }
        }

        private void SendCallback(IAsyncResult ar)
        {
            try
            {
                // Retrieve the socket from the state object.
                Socket client = (Socket)ar.AsyncState;

                // Complete sending the data to the remote device.
                int bytesSent = client.EndSend(ar);
                Console.WriteLine("Sent {0} bytes to server.", bytesSent);

                // Signal that all bytes have been sent.
                sendError = false;
                sendDone.Set();
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
                UpdateStatusDescription("At " + DateTime.Now + " Failed in SendCallback with error: " + e.ToString());
                sendError = true;
                sendDone.Set();
            }
        }

        // State object for receiving data from remote camera.
        private class StateObject
        {
            public Socket workSocket = null;    // clilent socket
            public byte[] buffer;               // receive buffer
            public int bytesRequested = 0;
            public int bytesReceived = 0;
        }

        // ManualResetEvent instances signal completion.
        private ManualResetEvent connectDone = new ManualResetEvent(false);
		private bool connectError = false;
        private bool sendError = false;
        private ManualResetEvent sendDone = new ManualResetEvent(false);
		private ManualResetEvent receiveDone = new ManualResetEvent(false);

		// The response from the remote device.
		private byte[] response;
		private byte[] prevResponse;

		private byte[] magicNumbers = new byte[] { 0x01, 0x00, 0x00, 0x00 };
		private byte[] huffman_table = null;

		private Socket client;


		private int[] huffman_table_int = new int[] {
			0xFF, 0xC4, 0x01, 0xA2, 0x00, 0x00, 0x01, 0x05, 0x01, 0x01,
			0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06,
			0x07, 0x08, 0x09, 0x0A, 0x0B, 0x01, 0x00, 0x03, 0x01,
			0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00,
			0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x03, 0x04,
			0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x10, 0x00,
			0x02, 0x01, 0x03, 0x03, 0x02, 0x04, 0x03, 0x05, 0x05,
			0x04, 0x04, 0x00, 0x00, 0x01, 0x7D, 0x01, 0x02, 0x03,
			0x00, 0x04, 0x11, 0x05, 0x12, 0x21, 0x31, 0x41, 0x06,
			0x13, 0x51, 0x61, 0x07, 0x22, 0x71, 0x14, 0x32, 0x81,
			0x91, 0xA1, 0x08, 0x23, 0x42, 0xB1, 0xC1, 0x15, 0x52,
			0xD1, 0xF0, 0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0A,
			0x16, 0x17, 0x18, 0x19, 0x1A, 0x25, 0x26, 0x27, 0x28,
			0x29, 0x2A, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A,
			0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x53,
			0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x63, 0x64,
			0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x73, 0x74, 0x75,
			0x76, 0x77, 0x78, 0x79, 0x7A, 0x83, 0x84, 0x85, 0x86,
			0x87, 0x88, 0x89, 0x8A, 0x92, 0x93, 0x94, 0x95, 0x96,
			0x97, 0x98, 0x99, 0x9A, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6,
			0xA7, 0xA8, 0xA9, 0xAA, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6,
			0xB7, 0xB8, 0xB9, 0xBA, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6,
			0xC7, 0xC8, 0xC9, 0xCA, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6,
			0xD7, 0xD8, 0xD9, 0xDA, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5,
			0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xF1, 0xF2, 0xF3, 0xF4,
			0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA, 0x11, 0x00, 0x02,
			0x01, 0x02, 0x04, 0x04, 0x03, 0x04, 0x07, 0x05, 0x04,
			0x04, 0x00, 0x01, 0x02, 0x77, 0x00, 0x01, 0x02, 0x03,
			0x11, 0x04, 0x05, 0x21, 0x31, 0x06, 0x12, 0x41, 0x51,
			0x07, 0x61, 0x71, 0x13, 0x22, 0x32, 0x81, 0x08, 0x14,
			0x42, 0x91, 0xA1, 0xB1, 0xC1, 0x09, 0x23, 0x33, 0x52,
			0xF0, 0x15, 0x62, 0x72, 0xD1, 0x0A, 0x16, 0x24, 0x34,
			0xE1, 0x25, 0xF1, 0x17, 0x18, 0x19, 0x1A, 0x26, 0x27,
			0x28, 0x29, 0x2A, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A,
			0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x53,
			0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x63, 0x64,
			0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x73, 0x74, 0x75,
			0x76, 0x77, 0x78, 0x79, 0x7A, 0x82, 0x83, 0x84, 0x85,
			0x86, 0x87, 0x88, 0x89, 0x8A, 0x92, 0x93, 0x94, 0x95,
			0x96, 0x97, 0x98, 0x99, 0x9A, 0xA2, 0xA3, 0xA4, 0xA5,
			0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xB2, 0xB3, 0xB4, 0xB5,
			0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xC2, 0xC3, 0xC4, 0xC5,
			0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xD2, 0xD3, 0xD4, 0xD5,
			0xD6, 0xD7, 0xD8, 0xD9, 0xDA, 0xE2, 0xE3, 0xE4, 0xE5,
			0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xF2, 0xF3, 0xF4, 0xF5,
			0xF6, 0xF7, 0xF8, 0xF9, 0xFA
		};
	}
}
