﻿// handle MJPEG video streaming cameras

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net.Sockets;
using System.Threading;
using System.Net;
using System.Collections;
using System.IO;
using System.Drawing;
using System.ComponentModel;

namespace ThunderDashboard
{
    partial class CameraControl : PictureBox
    {
        // runs the camera (for background worker) until told to exit camera
        private void RunCameraMpej()
        {
            string boundary = "";   // string that separates the images
            int timeout;
            WebResponse resp = null;

            // determine URL for camera feed (e.g. http://roborio-1511-frc.local:1181/stream.mjpg)
            string sourceURL = "http://" + _cameraIp + ":" + _port + "/stream.mjpg";

            // get connection to the robot stream
            while (!_shutdownCamera && !_resetCamera)
            {
                // Connect to remote camera.
                try
                {
                    // create HTTP request
                    HttpWebRequest req = HttpWebRequest.CreateHttp(sourceURL);
                    if (req == null)
                    {
                        // robot probably not ready just yet, so try again
                        UpdateStatusDescription("At " + DateTime.Now + " WebRequest failed");
                        return;
                    }
                    timeout = req.Timeout;      // remember the original default timeout

                    // get response
                    req.Timeout = 2000; // need to give the robot time to open a camera link for us, and give us a response
                    resp = req.GetResponse();
                    if (resp == null)
                    {
                        // robot didn't have any type of response, so it must not be fully ready
                        UpdateStatusDescription("At " + DateTime.Now + " GetReponse failed");
                        return;
                    }

                    // pull out the boundary string
                    if (!resp.ContentType.Contains("="))
                    {
                        UpdateStatusDescription("At " + DateTime.Now + " = missing");
                        return;
                    }
                    boundary = resp.ContentType.Split(new string[] { "=" }, StringSplitOptions.None)[1];
                    if (!boundary.StartsWith("--"))
                    {
                        boundary = "--" + boundary;
                    }
                    break;
                } // try
                catch (WebException we)
                {
                    if (we.Status == WebExceptionStatus.Timeout)
                    {
                        UpdateStatusDescription("At " + DateTime.Now + " timeout; retrying");
                    }
                    else
                    {
                        UpdateStatusDescription("At " + DateTime.Now + " Failed while connecting to camera: " + we.ToString());
                        return;
                    }
                }
                catch (Exception e)
                {
                    UpdateStatusDescription("At " + DateTime.Now + " Failed while connecting to camera: " + e.ToString());
                    return;
                }
            }
            if (_shutdownCamera || _resetCamera)
            {
                // we broke out of above while loop because we need to restart/shutdown, so get out of here
                return;
            }

            // open stream so we can get the images
            try
            {
                _stream = resp.GetResponseStream();
                if (_stream == null)
                {
                    UpdateStatusDescription("At " + DateTime.Now + " no stream");
                    return;
                }
                _stream.ReadTimeout = 1000; // default is 300 seconds; WAY too long
            }
            catch (WebException we)
            {
                if (we.Status == WebExceptionStatus.Timeout)
                {
                    UpdateStatusDescription("At " + DateTime.Now + " timeout; retrying");
                }
                else
                {
                    UpdateStatusDescription("At " + DateTime.Now + " Failed getting stream: " + we.ToString());
                    return;
                }
            }
            catch (Exception e)
            {
                UpdateStatusDescription("At " + DateTime.Now + " Failed to get stream: " + e.ToString());
                return;
            }

            // now that the stream is open, get each image until we are told to shutdown/restart
            try
            {
                _totalInBuffer = 0;
                while (!_shutdownCamera && !_resetCamera)
                {
                    // find the boundary
                    if (!FindBoundary(boundary))
                    {
                        break;
                    }

                    // get image's length and make sure we have room for it
                    int imageLength = GetImageLength();
                    if (imageLength <= 0)
                    {
                        if (imageLength == 0) { UpdateStatusDescription("At " + DateTime.Now + " bad header"); }
                        break;
                    }
                    if (imageLength > _buffer.Length)
                    {
                        // don't have neough space to hold the image
                        UpdateStatusDescription("At " + DateTime.Now + " image too big " + imageLength.ToString());
                        break;
                    }

                    // get bytes for the image, until we are told to shutdown/restart
                    DateTime timeEnd = DateTime.Now + TimeSpan.FromMilliseconds(2000);
                    int howMuchToRead = imageLength - _totalInBuffer;
                    while ((howMuchToRead > 0) && !_shutdownCamera && !_resetCamera)
                    {
                        if (timeEnd.CompareTo(DateTime.Now) < 0)
                        {
                            _resetCamera = true;
                            break;
                        }
                        int read = _stream.Read(_buffer, _totalInBuffer, howMuchToRead);
                        howMuchToRead -= read;
                        _totalInBuffer += read;
                    }
                    if (_totalInBuffer < imageLength)
                    {
                        break;  // didn't get enough bytes; probably told to shutdown/restart
                    }

                    // get bitmap to send to UI
                    byte[] processedImage = new byte[imageLength];
                    if (processedImage != null)
                    {
                        Buffer.BlockCopy(_buffer, 0, processedImage, 0, imageLength);

                        UpdateStatusDescription("At " + DateTime.Now + " got image");
                        //File.WriteAllBytes(@"C:\Users\Robotics\Documents\testimage.jpg", processedImage);

                        // make sure the image is updated in the UI thread
                        Invoke(new UpdateImageDelegate(UpdateImage), processedImage);
                    }

                    // remove image from the buffer
                    _totalInBuffer -= imageLength;
                    Buffer.BlockCopy(_buffer, imageLength, _buffer, 0, _totalInBuffer);
                }
            }
            catch (WebException we)
            {
                if (we.Status == WebExceptionStatus.Timeout)
                {
                    UpdateStatusDescription("At " + DateTime.Now + " timeout; retrying");
                }
                else
                {
                    UpdateStatusDescription("At " + DateTime.Now + " Failed getting images: " + we.ToString());
                    return;
                }
            }
            catch (Exception e)
            {
                UpdateStatusDescription("At " + DateTime.Now + " Failed reading images: " + e.ToString());
            }

            // make sure the stream is closed
            if (_stream != null)
            {
                _stream.Close();
                _stream = null;
            }
        }

        // find the boundary line in the stream
        // returns: true if boundary was found, false if never found
        private bool FindBoundary(string boundary)
        {
            // keep looking until we are told to shutdown/restart
            while (!_shutdownCamera && !_resetCamera)
            {
                // found it?
                if (ReadLine() == boundary)
                {
                    return (true);
                }
            }
            return (false);
        }

        // get the length of the image data from the image header info
        // return: length of the image data, or 0 is bad header, or -1 it shutdown/restart
        private int GetImageLength()
        {
            bool imageJpeg = false;
            int imageLength = 0;
            string line;

            // go until we have a blank line (that indicates the end of the iamge header, data follows right after it)
            while (!_shutdownCamera && !_resetCamera)
            {
                // get a line and make it lower case (we need case-insensitive compares)
                line = ReadLine().ToLowerInvariant();

                // look for the info that we need; as well as being at the end of the header
                if (line == "")                                 // at end of header?
                {
                    return (imageJpeg ? imageLength : 0);       // make sure content was good
                }
                else if (line.StartsWith("content-type: "))     // info about the type of image it is?
                {
                    imageJpeg = line.Contains("image/jpeg");    // we only want JPEG
                }
                else if (line.StartsWith("content-length: "))   // info about the length of the image
                {
                    try
                    {
                        // convert length
                        imageLength = Convert.ToInt32(line.Substring(16));
                    }
                    catch
                    {
                        imageLength = 0;
                    }
                }
            }

            return (-1);
        }

        // read a line from the stram
        private byte _carriageReturn = Convert.ToByte('\r');
        private byte _lineFeed = Convert.ToByte('\n');
        private string ReadLine()
        {
            try
            {
                // make sure we have at least a line's worth of data.
                // we pick 64 because no line will be longer than, and any image
                // will be more than because it will at least have a JPEG header.
                DateTime timeEnd = DateTime.Now + TimeSpan.FromMilliseconds(1000);
                while (_totalInBuffer < 64)
                {
                    // see if we need to get out of here, because we need to end the feed
                    if (_shutdownCamera || _resetCamera)
                    {
                        return ("");
                    }

                    // see if it took too long
                    if (timeEnd.CompareTo(DateTime.Now) < 0)
                    {
                        _resetCamera = true;
                        return ("");
                    }

                    // let's read more than we will need, to be efficient
                    int read = _stream.Read(_buffer, _totalInBuffer, 1000);
                    _totalInBuffer += read;
                }

                // find the end of the line (do 1 less than buffer because we are checking for 2 bytes)
                for (int i = 0; i < _totalInBuffer - 1; i++)
                {
                    // find the 
                    if ((_buffer[i] == _carriageReturn) && (_buffer[i + 1] == _lineFeed))
                    {
                        // convert the line to string
                        string line = Encoding.ASCII.GetString(_buffer, 0, i);

                        // remove the line from the read in data
                        _totalInBuffer -= (i + 2);
                        Buffer.BlockCopy(_buffer, (i + 2), _buffer, 0, _totalInBuffer);
                        return (line);
                    }
                }
            }
            catch
            {
                // we do not care why
            }

            // there is no line, or the line is longer than 64 characters (which is bad)
            return ("");
        }



#if WANT_SINGLE
        // runs the camera (for background worker) until told to exit camera
        // Returns: true if we should wait and retry again
        private bool RunCameraMpejSingle()
        {
            bool ret = false;

            // determine URL for camera feed (e.g. http://roborio-1511-frc.local:1181/stream.mjpg)
            string sourceURL = "http://" + _cameraIp + ":" + _port + "/stream.mjpg";

            while (!_shutdownCamera && !_resetCamera)		// needed incase we are stuck in outer try/catch
            {
                // Connect to remote camera.
                try
                {
                    // create HTTP request
                    HttpWebRequest req = HttpWebRequest.CreateHttp(sourceURL);
                    if (req == null)
                    {
                        UpdateStatusDescription("At " + DateTime.Now + " WebRequest failed");
                        return (true);
                    }
                    req.Timeout = 2000;

                    // get response
                    WebResponse resp = req.GetResponse();
                    if (resp == null)
                    {
                        UpdateStatusDescription("At " + DateTime.Now + " GetReponse failed");
                        return (true);
                    }

                    // pull out the boundary string
                    if (!resp.ContentType.Contains("="))
                    {
                        UpdateStatusDescription("At " + DateTime.Now + " = missing");
                        return (true);
                    }
                    string boundary = resp.ContentType.Split(new string[] { "=" }, StringSplitOptions.None)[1];
                    if (!boundary.StartsWith("--"))
                    {
                        boundary = "--" + boundary;
                    }
                    boundary += "\r\n";


                    byte[] boundaryBytes = Encoding.ASCII.GetBytes(boundary);
                    // data between boundary and content length
                    // data between content length and 
                    byte[] contentTypeBytes = Encoding.ASCII.GetBytes("Content-Type: image/jpeg\r\n");
                    byte[] contentLengthBytes = Encoding.ASCII.GetBytes("Content-Length: ");
                    byte[] timestampBytes = Encoding.ASCII.GetBytes("X-Timestamp: ");
                    byte[] newlineBytes = Encoding.ASCII.GetBytes("\r\n");

                    // get response stream
                    _stream = resp.GetResponseStream();
                    if (_stream == null)
                    {
                        return (true);
                    }
                    _totalInBuffer = 0;

                    // find the header content before the image
                    if (!FindString(boundaryBytes, boundaryBytes.Length, true))
                    {
                        _stream.Close();
                        _stream = null;
                        UpdateStatusDescription("At " + DateTime.Now + " boundary missing");
                        return (true);
                    }
                    int headerInfo = boundaryBytes.Length;
                    if (!FindString(contentTypeBytes, contentTypeBytes.Length))
                    {
                        _stream.Close();
                        _stream = null;
                        UpdateStatusDescription("At " + DateTime.Now + " Content-Type missing");
                        return (true);
                    }
                    headerInfo += contentTypeBytes.Length;
                    if (!FindString(contentLengthBytes, contentLengthBytes.Length))
                    {
                        _stream.Close();
                        _stream = null;
                        UpdateStatusDescription("At " + DateTime.Now + " Content-Length missing");
                        return (true);
                    }
                    headerInfo += contentLengthBytes.Length;
                    int imageLength = 0;
                    int i = 0;
                    while ((_buffer[i] >= (byte)0x30) && (_buffer[i] <= (byte)0x39))
                    {
                        imageLength *= 10;
                        imageLength += Convert.ToInt32(_buffer[i]) - 0x30;
                        i++;
                    }
                    if (_buffer[i] != '\r')
                    {
                        _stream.Close();
                        _stream = null;
                        UpdateStatusDescription("At " + DateTime.Now + " newline missing (1)");
                        return (true);
                    }
                    i++;
                    if (_buffer[i] != '\n')
                    {
                        _stream.Close();
                        _stream = null;
                        UpdateStatusDescription("At " + DateTime.Now + " newline missing (2)");
                        return (true);
                    }
                    i++;
                    _totalInBuffer -= i;
                    Buffer.BlockCopy(_buffer, i, _buffer, 0, _totalInBuffer);
                    if (imageLength > _buffer.Length)
                    {
                        // don't have neough space to hold the image
                        _stream.Close();
                        _stream = null;
                        UpdateStatusDescription("At " + DateTime.Now + " newline missing (3)");
                        return (true);
                    }
                    headerInfo += i;

                    // 
                    if (!FindString(timestampBytes, timestampBytes.Length))
                    {
                        _stream.Close();
                        _stream = null;
                        UpdateStatusDescription("At " + DateTime.Now + " timestamp missing (1)");
                        return (true);
                    }
                    if (!FindString(newlineBytes, newlineBytes.Length, true))
                    {
                        _stream.Close();
                        _stream = null;
                        UpdateStatusDescription("At " + DateTime.Now + " timestamp missing (2)");
                        return (true);
                    }
                    if (!FindString(newlineBytes, newlineBytes.Length, true))
                    {
                        _stream.Close();
                        _stream = null;
                        UpdateStatusDescription("At " + DateTime.Now + " newline missing (5)");
                        return (true);
                    }

                    // get bytes until we have enough for the image
                    int howMuchToRead = imageLength - _totalInBuffer;
                    while (howMuchToRead > 0)
                    {
                        int read = _stream.Read(_buffer, _totalInBuffer, howMuchToRead);
                        howMuchToRead -= read;
                        _totalInBuffer += read;
                    }

                    // get bitmap to send to UI
                    byte[] processedImage = new byte[imageLength];
                    if (processedImage != null)
                    {
                        Buffer.BlockCopy(_buffer, 0, processedImage, 0, imageLength);

                        UpdateStatusDescription("At " + DateTime.Now + " got image");
                        //File.WriteAllBytes(@"C:\Users\Robotics\Documents\testimage.jpg", processedImage);

                        // make sure the image is updated in the UI thread
                        Invoke(new UpdateImageDelegate(UpdateImage), processedImage);
                    }

                    _totalInBuffer -= imageLength;
                    Buffer.BlockCopy(_buffer, imageLength, _buffer, 0, _totalInBuffer);

                    _stream.Close();
                    _stream = null;
                } // try
                catch (WebException we)
                {
                    if (_stream != null)
                    {
                        _stream.Close();
                        _stream = null;
                    }

                    if (we.Status == WebExceptionStatus.Timeout)
                    {
                        UpdateStatusDescription("At " + DateTime.Now + " timeout");
                    }
                    else
                    {
                        UpdateStatusDescription("At " + DateTime.Now + " Failed while connecting to camera: " + we.ToString());

                        // break out and start all over looking for the robot
                        break;
                    }
                }
                catch (Exception e)
                {
                    //Console.WriteLine(e.ToString());
                    UpdateStatusDescription("At " + DateTime.Now + " Failed while connecting to camera: " + e.ToString());

                    if (_stream != null)
                    {
                        _stream.Close();
                        _stream = null;
                    }

                    // break out and start all over looking for the robot
                    break;
                }
            }
            UpdateStatusDescription("At " + DateTime.Now + " DEAD");

            return (ret);
        }


        private bool FindString(byte[] data, int length, bool goUntilFound = false)
        {
            int read;
            bool readyToCompare = false;

            // go until we have something to compare
            while (!readyToCompare)
            {
                // read until we have enough data
                while (_totalInBuffer < length)       // needed incase we are stuck in outer try/catch
                {
                    // see if we need to get out of here
                    if (_shutdownCamera || _resetCamera)
                    {
                        return (false);
                    }

                    read = _stream.Read(_buffer, _totalInBuffer, 1000);
                    _totalInBuffer += read;
                }

                // if we do not need to find it (i.e. just go with what is in the buffer), then break out and go with what we already have
                if (!goUntilFound)
                {
                    readyToCompare = true;
                    break;
                }

                // make sure the buffer starts with the begining of what we are looking for
                bool foundStart = true;
                for (int i = 0; i < _totalInBuffer; i++)
                {
                    // find the first character
                    if (_buffer[i] == data[0])
                    {
                        _totalInBuffer -= i;
                        Buffer.BlockCopy(_buffer, i, _buffer, 0, _totalInBuffer);

                        // if we still have enough data, then we can break out and compare
                        if (_totalInBuffer >= length)
                        {
                            readyToCompare = true;
                        }
                        foundStart = true;
                        break;
                    }
                }
                if (!foundStart)
                {
                    // no beginning byte found, then forget about all the data we have and start over
                    _totalInBuffer = 0;
                }
            }

            // see if it compares
            if (BufferCompare(_buffer, data, length))
            {
                _totalInBuffer -= length;
                Buffer.BlockCopy(_buffer, length, _buffer, 0, _totalInBuffer);
                return (true);
            }

            return (false);
        }



        // compare two byte arrays to see if they are equal
        //
        // parameters:
        //    data1        - first buffer to compare
        //    data2        - second buffer to compare
        //    length       - number of bytes to compare
        //
        // returns true if the buffers are equal
        private bool BufferCompare(byte[] data1, byte[] data2, int length)
        {
            // make the buffer start at what could be the "boundary"
            for (int i = 0; i < length; i++)
            {
                if (data1[i] != data2[i])
                {
                    return (false);
                }
            }
            return (true);
        }
#endif



        // member variables
        private byte[] _buffer = new byte[(640 * 480 * 3) + 65536]; // hold our worst case image size, plus a little
        private int _totalInBuffer = 0; // total number of bytes in _buffer that are 'valid'
        private Stream _stream = null; // stream being used to read in image
    }
}
