﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net.Sockets;
using System.Threading;
using System.Net;
using System.Collections;
using System.IO;
using System.Drawing;
using System.ComponentModel;

namespace ThunderDashboard
{
    public partial class CameraControl : PictureBox
    {
        // delegate that gets called when image has been processed
        // targetInfo - each target's info
        // targets    - how many targets are in targetInfo
        // imageWidth - the width of the entire image
        // imageHeight - the height of the entire image
        //
        // caller returns true if they want the targets to be drawn
        public delegate bool VisionDelegate(ThunderVision.TARGET_INFO[] targetInfo, int targets, int imageWidth, int imageHeight);

        // camera resolution
        public enum Resolution
        {
            ResolutionDoNotSet = 99,
            Resolution640x480 = 0,
            Resolution320x240 = 1,
            Resolution160x120 = 2,
        };

        private string FPS(int fps)
        {
            return (string.Format("{0} fps", fps));
        }

        // constructor
        public CameraControl()
        {
            _started = false;

            // create control for resolutions
            _resChoices = new ComboBox();
            _resChoices.DropDownStyle = ComboBoxStyle.DropDownList;
            _resChoices.SelectedIndexChanged += _resChoices_SelectedIndexChanged;
            _resChoices.Size = new Size(76, 21);
            _resChoices.Location = new Point(0, Bottom - _resChoices.Height);
            _resChoices.Anchor = AnchorStyles.Left | AnchorStyles.Bottom;
            _resChoices.Items.Add("640 x 480");
            _resChoices.Items.Add("320 x 240");
            _resChoices.Items.Add("160 x 120");
            AddToolTip(_resChoices, "Resolution");
            Controls.Add(_resChoices);

            // create control for frames per second
            _fpsChoices = CreateToolbarDropdown(null, _fpsChoices_SelectedIndexChanged, "Frame Rate", new string[] { FPS(30), FPS(25), FPS(20), FPS(15), FPS(10), FPS(5) }, FPS(Properties.Settings.Default.CameraFrameRate), 56);
            _fpsChoices.Location = new Point(Right - _fpsChoices.Width, Bottom - _fpsChoices.Height);
            _fpsChoices.Anchor = AnchorStyles.Right | AnchorStyles.Bottom;
            Controls.Add(_fpsChoices);

            // add brightness and exposure panel
            _panelBright = new TableLayoutPanel();
            _panelBright.RowStyles.Add(new RowStyle(SizeType.AutoSize));
            _panelBright.Anchor = AnchorStyles.Right | AnchorStyles.Top;
            _panelBright.AutoSize = true;
            _panelBright.AutoSizeMode = AutoSizeMode.GrowAndShrink;
            _panelBright.BackColor = Color.FromArgb(207, 214, 229);
            _panelBright.Padding = new Padding(0);
            _panelBright.Margin = new Padding(0);
            _defaults = CreateToolbarButton(_panelBright, Properties.Resources.Defaults, _defaults_Click, "Set 2017 defaults");
            //CreateToolbarText(_panelBright, "B:");
            _brightness = CreateToolbarNumericUpDown(_panelBright, _brightness_ValueChanged, "Brightness", 0, 100, Properties.Settings.Default.CameraBrightness);
            //CreateToolbarText(_panelBright, "E:");
            _exposure = CreateToolbarNumericUpDown(_panelBright, _exposure_ValueChanged, "Exposure", -1, 100, Properties.Settings.Default.CameraExposure);
            _whiteBalance = CreateToolbarDropdown(_panelBright, _whiteBalance_SelectedIndexChanged, "White Balance", new string[] { "auto", "fi", "fo1", "fo2", "ff1", "ff2" }, Properties.Settings.Default.CameraWhiteBalance, 44);
            _panelBright.Location = new Point(Right - _panelBright.Width, 0);
            Controls.Add(_panelBright);

            // add image toolbar panel
            _panelImage = new TableLayoutPanel();
            _panelImage.AutoSize = true;
            _panelImage.AutoSizeMode = AutoSizeMode.GrowAndShrink;
            _panelImage.BackColor = Color.FromArgb(207, 214, 229);
            _panelImage.RowStyles.Add(new RowStyle(SizeType.AutoSize));
            _panelImage.Anchor = AnchorStyles.Left | AnchorStyles.Top;
            _panelImage.Padding = new Padding(0);
            _playButton = CreateToolbarButton(_panelImage, Properties.Resources.Pause, _playButton_Click, "Pause live stream", true);
            _enableButton = CreateToolbarButton(_panelImage, Properties.Resources.CameraOff, _enableButton_Click, "Stop camera connection", true);
            _liveProcessButton = CreateToolbarButton(_panelImage, Properties.Resources.LiveProcessOn, _liveProcessButton_Click, "Start live processing, even if robot isn't asking", true);
            _saveButton = CreateToolbarButton(_panelImage, Properties.Resources.Save, _saveButton_Click, "Save current image to a bitmap file");
            _loadButton = CreateToolbarButton(_panelImage, Properties.Resources.Open, _loadButton_Click, "Load bitmap file to be current image");
            _processButton = CreateToolbarButton(_panelImage, Properties.Resources.Process, _processButton_Click, "Process currently displayed image", true);
            _panelImage.Location = new Point(0, 0);
            Controls.Add(_panelImage);

            SetPlayMode(true);
        }


        private Button CreateToolbarButton(TableLayoutPanel panel, Image image, EventHandler handler, string toolTip, bool separator = false)
        {
            Button control = new Button();
            control.AutoSize = true;
            control.AutoSizeMode = AutoSizeMode.GrowAndShrink;
            control.ImageAlign = ContentAlignment.MiddleCenter;
            control.FlatStyle = FlatStyle.Flat;
            control.FlatAppearance.BorderSize = 0;
            control.Margin = new Padding(0, 0, (separator ? 10 : 2), 0);
            control.Image = image;
            control.Click += handler;


            if (panel != null)
            {
                panel.ColumnStyles.Add(new ColumnStyle(SizeType.AutoSize));
                panel.Controls.Add(control, panel.Controls.Count, 0);
            }

            AddToolTip(control, toolTip);
            return (control);
        }
        private NumericUpDown CreateToolbarNumericUpDown(TableLayoutPanel panel, EventHandler handler, string toolTip, decimal min = 0, decimal max = 100, decimal value = 100, int width = 40)
        {
            NumericUpDown control = new NumericUpDown();
            control.Minimum = min;
            control.Maximum = max;
            control.Margin = new Padding(0, 0, 2, 0);
            control.Value = value;
            control.Width = width;
            control.ValueChanged += handler;


            if (panel != null)
            {
                panel.ColumnStyles.Add(new ColumnStyle(SizeType.AutoSize));
                panel.Controls.Add(control, panel.Controls.Count, 0);
            }

            AddToolTip(control, toolTip);
            return (control);
        }
        private Label CreateToolbarText(TableLayoutPanel panel, string text)
        {
            Label control = new Label();
            control.AutoSize = true;
            control.Margin = new Padding(0, 2, 0, 0);
            control.Text = text;


            if (panel != null)
            {
                panel.ColumnStyles.Add(new ColumnStyle(SizeType.AutoSize));
                panel.Controls.Add(control, panel.Controls.Count, 0);
            }

            //AddToolTip(control, toolTip);
            return (control);
        }
        private ComboBox CreateToolbarDropdown(TableLayoutPanel panel, EventHandler handler, string toolTip, string[] list, string value, int width = 56)
        {
            ComboBox control = new ComboBox();
            control.DropDownStyle = ComboBoxStyle.DropDownList;
            control.Margin = new Padding(0, 0, 2, 0);
            control.SelectedIndexChanged += handler;
            control.Size = new Size(width, 21);
            foreach (string item in list)
            {
                control.Items.Add(item);
                if (item == value)
                {
                    control.SelectedItem = value;
                }
            }

            if (panel != null)
            {
                panel.ColumnStyles.Add(new ColumnStyle(SizeType.AutoSize));
                panel.Controls.Add(control, panel.Controls.Count, 0);
            }

            AddToolTip(control, toolTip);
            return (control);
        }
        private void AddToolTip(Control control, string toolTip)
        {
            // set tool tip
            if (!string.IsNullOrEmpty(toolTip))
            {
                if (_tooltip == null)
                {
                    _tooltip = new ToolTip();
                    _tooltip.ShowAlways = true;
                }
                _tooltip.SetToolTip(control, toolTip);
            }
        }

        // load image
        private void _loadButton_Click(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.CheckFileExists = true;
            ofd.CheckPathExists = true;
            ofd.DefaultExt = ".bmp";
            ofd.Filter = "Bitmap Files(*.BMP)| *.BMP";
            //ofd.InitialDirectory = "C:\\Users\\Robotics\\Desktop";
            if (DialogResult.OK == ofd.ShowDialog())
            {
                try
                {
                    _loadedFile = ofd.FileName;
                    this.Image = new Bitmap(ofd.FileName);
                }
                catch
                {
                    MessageBox.Show("Unable to open the image");
                }
            }
        }
        // enable/disable use of the camera
        private void _enableButton_Click(object sender, EventArgs e)
        {
            _cameraEnabled = !_cameraEnabled;
            _enableButton.Image = (_cameraEnabled ? Properties.Resources.CameraOff : Properties.Resources.CameraOn);
            _tooltip.SetToolTip(_enableButton, (_cameraEnabled ? "Stop camera connection" : "Start camera connection"));
            if (_cameraEnabled)
            {
                this.Image = null;  // remove 'disabled' image
                Start();
            }
            else
            {
                this.Image = Properties.Resources.CameraDisabled;
                Stop();
            }
        }
        // play/pause
        private void _playButton_Click(object sender, EventArgs e)
        {
            SetPlayMode(!_updateDisplay);
        }
        private void SetPlayMode(bool play)
        {
            _updateDisplay = play;
            _playButton.Image = (_updateDisplay ? Properties.Resources.Pause : Properties.Resources.Play);
            _tooltip.SetToolTip(_playButton, (_updateDisplay ? "Pause live stream" : "Restart live stream"));
            _saveButton.Enabled = !_updateDisplay;
            _loadButton.Enabled = !_updateDisplay;
            _processButton.Enabled = !_updateDisplay;
            _liveProcessButton.Enabled = _updateDisplay;
        }
        // save
        private void _saveButton_Click(object sender, EventArgs e)
        {
            SaveFileDialog sfd = new SaveFileDialog();
            sfd.AddExtension = true;
            sfd.CheckPathExists = true;
            sfd.OverwritePrompt = true;
            sfd.CheckPathExists = true;
            sfd.DefaultExt = ".bmp";
            sfd.Filter = "Bitmap Files(*.BMP)| *.BMP";
            //sfd.InitialDirectory = "C:\\Users\\Robotics\\Desktop";
            if (DialogResult.OK == sfd.ShowDialog())
            {
                try
                {
                    this.Image.Save(sfd.FileName, System.Drawing.Imaging.ImageFormat.Bmp);
                }
                catch
                {
                    MessageBox.Show("Unable to save the image");
                }
            }
        }
        // process image
        private void _processButton_Click(object sender, EventArgs e)
        {
            //if (!string.IsNullOrEmpty(_file))
            {
                ProcessTarget(_loadedFile);
            }
        }
        // live processing of images
        private void _liveProcessButton_Click(object sender, EventArgs e)
        {
            _liveProcessing = !_liveProcessing;
            _liveProcessButton.Image = (_liveProcessing ? Properties.Resources.LiveProcessOff : Properties.Resources.LiveProcessOn);
            _tooltip.SetToolTip(_liveProcessButton, (_liveProcessing ? "Stop live processing, only do it when robot asks" : "Start live processing, even if robot isn't asking"));
        }

        // start and stop vision processing
        public void VisionStart()
        {
            _visionProcessing = (_visionDelegate != null);
        }
        public void VisionStop()
        {
            _visionProcessing = false;
        }

        // start camera processing. this should stay alive until we are closing the dashboard, or user stopped the feed
        public void Start()
        {
            Start(_cameraIp, _port, _statusTextBox, _cameraFlip, _visionDelegate);
        }
        public void Start(string ipAddress, int port /*= 1180*/, TextBox status /*= null*/, bool flip /*= false*/, VisionDelegate visionDelegate)
        {
            // if camera isn't enabled, then we cannot start it
            if (!_cameraEnabled)
            {
                return;
            }

            // hide/show vision processing controls based on whether it is supported or not
            _liveProcessButton.Visible = (visionDelegate != null);
            _loadButton.Visible = (visionDelegate != null);
            _processButton.Visible = (visionDelegate != null);

            // if camera was already started, then let's kill that previous instance first
            if (_started)
            {
                Stop();
                Thread.Sleep(250); // wait for current camera to stop
                _started = false;
            }

            // save our settings
            _port = port;
            _cameraIp = ipAddress;
            _cameraFlip = flip;
            _statusTextBox = status;
            _visionDelegate = visionDelegate;

            // get resolution and frame rate
            try
            {
                _resChoices.SelectedIndex = Properties.Settings.Default.CameraResolution;
                _fpsChoices.SelectedItem = FPS(Properties.Settings.Default.CameraFrameRate);
                _brightness.Value = Properties.Settings.Default.CameraBrightness;
                _exposure.Value = Properties.Settings.Default.CameraExposure;
            }
            catch
            {
                _resChoices.SelectedIndex = 0;
                _fpsChoices.SelectedIndex = 0;
                _brightness.Value = 50;
                _exposure.Value = 50;
            }
            UpdateRobot();  // make sure the robot is in sync

            // get the camera thread running
            MakeSureIsRunning();
        }

        // reset the cammera (close thread and restart it)
        public void Restart()
        {
            // handle it if camera thread is already running
            if (_bw != null)
            {
                // if thread was started, then tell it to reset
                if (_started)
                {
                    UpdateStatusDescription("restarting");
                    _resetCamera = true;
                    return;
                }
            }

            // camera thread not running, so get it started
            MakeSureIsRunning();
        }

        private void MakeSureIsRunning()
        {
            // if already running, then nothing to do
            if (_started)
            {
                return;
            }

            // only create background worker if not already running
            if (_bw == null)
            {
                _bw = new BackgroundWorker();
                _bw.DoWork += new DoWorkEventHandler(StartConnection);
            }

            // if it isn't running, then start it up
            if (!_bw.IsBusy)
            {
                _bw.RunWorkerAsync();
            }

            _started = true;
        }

        // called when the network tables is disconnected
        public void Disconnected()
        {
            // restart the camera and have it re-start looking for it
            _resetCamera = true;
        }

        // called when network tables are re-connected
        public void Connected()
        {
            // since we just got the network tables, let's make sure our settings are there for the robot to see
            UpdateRobot();
        }

        // stop using camera. this should only be called when we are completely done with the camera
        public void Stop()
		{
            // if camera wasn't started, then we have nothing to stop
            if (!_started)
            {
                // if camera is disabled, then update button's image
                if (!_cameraEnabled)
                {
                    this.Image = Properties.Resources.CameraDisabled;
                }
                return;
            }

            _resetCamera = false; // do not restart the camera
            _shutdownCamera = true; // we are done with the camera
		}

        // set the resolution for the camera
        private void SetResolution(Resolution resolution)
        {
            // if the rate didn't change, then just return
            if (_resolution == resolution)
            {
                return;
            }

            // save the new value and refresh the camera
            _resolution = resolution;
            Restart();
        }

        // set the frames per second for the camera
        private void SetFrameRate(int fps)
        {
            // if the rate didn't change, then just return
            if (_fps == fps)
            {
                return;
            }

            // save the new value and refresh the camera
            _fps = fps;
            Restart();
        }

        // set all network table based values
        private void UpdateRobot()
        {
            DashboardNetworkTable.PutNumber("camera_brightness", (double)_brightness.Value);
            DashboardNetworkTable.PutNumber("camera_exposure", (double)_exposure.Value);
            if (_whiteBalance.SelectedIndex >= 0)
            {
                DashboardNetworkTable.PutString("camera_whitebalance", (_whiteBalance.SelectedItem as string));
            }
        }

        // background worker thread
        private void StartConnection(object obj, DoWorkEventArgs eEventArgs)
        {
            bool displayedWaitingMesage = false;

            // retry if we are reseting/changing settings
            do
            {
                // if we just reset, then we want to display the message
                if (_resetCamera)
                {
                    displayedWaitingMesage = false;
                }
                _resetCamera = false;   // clear flag so we only reset once

                // see if the robot is even there
                if (Program.PingRobot(_cameraIp))
                {
                    _shutdownCamera = false;        // init it once
                    displayedWaitingMesage = false; // since we doign to attempt to connect, we can display this again
                    UpdateStatusDescription("connecting...");	// init status window
                    if (_port == 1180)
                    {
                        RunCamera();
                    }
                    else
                    {
                        RunCameraMpej();
                    }
                }
                else if (!displayedWaitingMesage)   // no robot, so say we are waiting
                {
                    displayedWaitingMesage = true;  // only once while robot is missing
                    UpdateStatusDescription("waiting to retry camera connection");
                }

                // robot is not there, so wait 3 seconds and try again. but wait
                // in 100ms intervals in case user closes the form while we were waiting
                for (int i = 0; ((i < 30) && !_shutdownCamera && !_resetCamera); i++)
                {
                    Thread.Sleep(100);
                }

            } while (!_shutdownCamera);

            UpdateStatusDescription("camera ended..."); // init status window
            if (!_cameraEnabled)
            {
                // camera feed was disabled, so show image indicating such.
                // we do this here because an image might have come in when we shut it down
                if (InvokeRequired)
                {
                    BeginInvoke(new UpdateImageDelegate(UpdateImage), new byte[] { 0 });
                }
                else
                {
                    UpdateImage(null);
                }
            }
        }

        private Bitmap byteArrayToImage(byte[] byteArrayIn)
        {
            // make sure we got something
            if (byteArrayIn.Length < 1)
            {
                return (null);
            }
            try
            {
                MemoryStream ms = new MemoryStream(byteArrayIn);
                Bitmap returnImage = new Bitmap(ms);
                //Console.WriteLine("Created Image, size={0}", Image.Size);
                return returnImage;
            }
            catch
            {
                return (null);
            }
        }

        delegate void UpdateStatusDescriptionDelegate(string message);

		private void UpdateStatusDescription(string message)
		{
			if (InvokeRequired)
			{
				BeginInvoke(new UpdateStatusDescriptionDelegate(UpdateStatusDescription), message);
				return;
			}

			if (_statusTextBox != null)
			{
				if (message.StartsWith("At") || (message == ""))
				{
					_statusTextBox.Text = message;
				}
				else
				{
					_statusTextBox.Text = message + "\r\n" + _statusTextBox.Text;
				}
			}
        }

        //MAB - added so the image update happens in the correct thread
        delegate void UpdateImageDelegate(byte[] processedImage);
        private void UpdateImage(byte[] processedImage)
        {
            if ((processedImage == null) || (processedImage.Count() == 1))
            {
                if (!_cameraEnabled)
                {
                    this.Image = Properties.Resources.CameraDisabled;
                }
                _loadedFile = null; // image is not from the user anymore
                return;
            }

            // id display is paused, then do not show the new image
            if (!_updateDisplay)
            {
                return;
            }

            if (_cameraFlip)
            {
                Bitmap image = byteArrayToImage(processedImage);
                image.RotateFlip(RotateFlipType.Rotate180FlipNone);
                this.Image = image;
            }
            else
            {
                this.Image = byteArrayToImage(processedImage);
            }
            _loadedFile = null; // image is not from the user anymore

            // if we shouldn't do processing, then skip processing.
            // NOTE we do this here so ProcesTarget can be called when there is no robot
            //      and the user loaded their own image; plus it saves calling the method
            if (_visionProcessing || _liveProcessing)
            {
                ProcessTarget();
            }
        }

        // process targets using the currently displayed image
        // file - is supplied is full path of the file name to use for debug images.
        //        set to null so no debug images are created
        private void ProcessTarget(string file = null)
        {
            // valid the image
            if ((this.Image == null) || !(this.Image is Bitmap))
            {
                return;
            }

            int width = this.Image.Width;
            int height = this.Image.Height;
            Bitmap bitmapSrc = this.Image as Bitmap;

            // validate the bitmap
            if ((bitmapSrc == null) || (bitmapSrc.PixelFormat != System.Drawing.Imaging.PixelFormat.Format24bppRgb))
            {
                return;
            }

            // if we didn't create the memory yet OR it is too small, then reallocate it
            // NOTE: we do it this way to save memory allocation time for each image we process
            if ((_rgbValue == null) || (_rgbValue.Length < (width * height)))
            {
                _rgbValue = null;
                _rgbValue = new ThunderVision.RGBValue[width * height];
            }

            // make a copy of the displayed image so we can lock its bits and change them
            // if we need to. we cannot locke the bits of what is being displayed because it messes
            // up what is being shown (i.e. get red X).
            // we need to make a Clone so that the PixelFormat is left untouched
            Rectangle rect = new Rectangle(0, 0, bitmapSrc.Width, bitmapSrc.Height);
            Bitmap bitmap = bitmapSrc.Clone(rect, bitmapSrc.PixelFormat);

            // lock the entire bitmap
            System.Drawing.Imaging.BitmapData bitmapData = bitmap.LockBits(rect, System.Drawing.Imaging.ImageLockMode.ReadWrite, bitmap.PixelFormat);

            // make sure we have enough memory to hold the image
            // NOTE: we do it this way to save memory allocation time for each image we process
            int bytes = Math.Abs(bitmapData.Stride) * bitmap.Height;
            if ((_bitmapRGB == null) || (_bitmapRGB.Length < bytes))
            {
                _bitmapRGB = null;
                _bitmapRGB = new byte[bytes];
            }

            // get the RGB values into an array
            System.Runtime.InteropServices.Marshal.Copy(bitmapData.Scan0, _bitmapRGB, 0, bytes);

            // pull out the RGB values for the target processing and do the processing
            // NOTE: values are in BGR order
            int dst = 0;
            int src;
            for (int h = 0; h < bitmap.Height; h++)
            {
                src = (h * bitmapData.Stride);
                for (int w = 0; w < bitmap.Width; w++)
                {
                    _rgbValue[dst].B = _bitmapRGB[src++];
                    _rgbValue[dst].G = _bitmapRGB[src++];
                    _rgbValue[dst].R = _bitmapRGB[src++];
                    dst++;
                }
            }
            _thunderVision.ProcessImage(_rgbValue, width, height, file);

            // make sure we have someone to call
            // NOTE: if not, we still draw the targets because were called manually by the user)
            if (_visionDelegate == null)
            {
                bitmap.UnlockBits(bitmapData);  // done with bitmap, we are leaving it alone
                bitmap = null;
                return;
            }

            // call delegate
            bool isValid = _visionDelegate(_thunderVision.targetInfo, _thunderVision.targets, width, height);

            // draw the targets on the image
            // NOTE: values are in BGR order
            int rowStride = width * 3;
            int pixel;
            byte r;
            byte g;
            byte b;
            for (int i = 0; i < _thunderVision.targets; i++)
            {
                ThunderVision.TARGET_INFO ti = _thunderVision.targetInfo[i];
                r = 255;
                if (isValid && ti.valid)
                {
                    g = b = 255;
                }
                else
                {
                    g = b = 0;
                }
                for (int x = ti.left; x <= ti.right; x++)           // draw top and bottom
                {
                    pixel = (ti.top * rowStride) + (x * 3);
                    _bitmapRGB[pixel++] = b; _bitmapRGB[pixel++] = g; _bitmapRGB[pixel++] = r;
                    pixel = (ti.bottom * rowStride) + (x * 3);
                    _bitmapRGB[pixel++] = b; _bitmapRGB[pixel++] = g; _bitmapRGB[pixel++] = r;
                }
                for (int y = ti.top; y <= ti.bottom; y++)           // draw left and right
                {
                    pixel = (y * rowStride) + (ti.left * 3);
                    _bitmapRGB[pixel++] = b; _bitmapRGB[pixel++] = g; _bitmapRGB[pixel++] = r;
                    pixel = (y * rowStride) + (ti.right * 3);
                    _bitmapRGB[pixel++] = b; _bitmapRGB[pixel++] = g; _bitmapRGB[pixel++] = r;
                }
            }

            // put back the updated values
            System.Runtime.InteropServices.Marshal.Copy(_bitmapRGB, 0, bitmapData.Scan0, bytes);

            // we are done updating the bitmap
            bitmap.UnlockBits(bitmapData);
            this.Image = bitmap;                // show the updated bitmap
        }

        // handle user changing resolution
        private void _resChoices_SelectedIndexChanged(object sender, EventArgs e)
        {
            int idx = ((_resChoices == null) ? -1 : _resChoices.SelectedIndex);

            switch (idx)
            {
                default:
                    return;     // unsupported value, so return now and do not save the setting
                case 0:
                    SetResolution(CameraControl.Resolution.Resolution640x480);
                    break;
                case 1:
                    SetResolution(CameraControl.Resolution.Resolution320x240);
                    break;
                case 2:
                    SetResolution(CameraControl.Resolution.Resolution160x120);
                    break;
            }

            // save the item
            try
            {
                Properties.Settings.Default.CameraResolution = idx;
                Properties.Settings.Default.Save();
            }
            catch
            {
            }
        }

        // handle user changing frame rate
        private void _fpsChoices_SelectedIndexChanged(object sender, EventArgs e)
        {
            // make sure something is selected
            int idx = ((_fpsChoices == null) ? -1 : _fpsChoices.SelectedIndex);
            if (idx < 0)
            {
                return;
            }
            try
            {
                int fps = Convert.ToInt32((_fpsChoices.SelectedItem as string).Split(' ')[0]);
                SetFrameRate(fps);

                // save the item
                Properties.Settings.Default.CameraFrameRate = fps;
                Properties.Settings.Default.Save();
            }
            catch
            {
            }
        }

        // handle brightness changes
        private void _brightness_ValueChanged(object sender, EventArgs e)
        {
            try
            {
                // make sure the robot knows
                UpdateRobot();

                // save the item
                Properties.Settings.Default.CameraBrightness = _brightness.Value;
                Properties.Settings.Default.Save();
            }
            catch
            {
            }
        }

        // handle exposure changes
        private void _exposure_ValueChanged(object sender, EventArgs e)
        {
            try
            {
                // make sure the robot knows
                UpdateRobot();

                // save the item
                Properties.Settings.Default.CameraExposure = _exposure.Value;
                Properties.Settings.Default.Save();
            }
            catch
            {
            }
        }

        // handle user changing white balance
        private void _whiteBalance_SelectedIndexChanged(object sender, EventArgs e)
        {
            // make sure something is selected
            int idx = ((_whiteBalance == null) ? -1 : _whiteBalance.SelectedIndex);
            if (idx < 0)
            {
                return;
            }
            try
            {
                // make sure the robot knows
                UpdateRobot();

                // save the item
                Properties.Settings.Default.CameraWhiteBalance = (_whiteBalance.SelectedItem as string);
                Properties.Settings.Default.Save();
            }
            catch
            {
            }
        }

        // camera default list
        private void _defaults_Click(object sender, EventArgs e)
        {
            if (DialogResult.Yes != MessageBox.Show("This will set the 2017 defaults.\n\nAre you sure?", "Set Camera Defaults", MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button2))
            {
                return;
            }
            _brightness.Value = 20;
            _exposure.Value = -1;
            _whiteBalance.SelectedItem = "ff1";
            _resChoices.SelectedItem = "320 x 240";
            _fpsChoices.SelectedItem = FPS(20);
        }
    
        private string _cameraIp;       // IP Address of camera
        private bool _cameraFlip;       // flip camera image?
		private int _port = 1180;       // port that has image
		private TextBox _statusTextBox; // where to put any status messages (for debugging
        private Resolution _resolution = Resolution.Resolution640x480; // camera resolution
        private int _fps = 30;          // camera frame rate per second
        private volatile bool _shutdownCamera = false;
		private BackgroundWorker _bw;
        private volatile bool _resetCamera = false;
        private volatile bool _started = false;

        // user controls
        private ComboBox _resChoices;                       // list of resolutions
        private ComboBox _fpsChoices;                       // list of frame rates
        private TableLayoutPanel _panelImage;               // panel to hold toolbar buttons for image
        private NumericUpDown _brightness;                  // camera brightness
        private NumericUpDown _exposure;                    // camera exposure
        private ComboBox _whiteBalance;                     // camera white balance
        private Button _defaults;                           // defaults
        private TableLayoutPanel _panelBright;              // panel to hold brightness and exposure
        private ToolTip _tooltip;                           // tooltips for all user controls
        private Button _loadButton;                         // load a bitmap into control
        private Button _enableButton = null;                // stop/start camera connection
        private Button _playButton = null;                  // pause/play live stream
        private Button _saveButton = null;                  // save image that is in control
        private Button _processButton = null;               // process image that is in control
        private Button _liveProcessButton = null;           // live process of streaming video
        private bool _cameraEnabled = true;                 // stop/start camera connection
        private bool _updateDisplay = true;                 // update image when it comes from robot?
        private string _loadedFile = null;                  // file loaded by user
        private bool _liveProcessing = false;               // do live processing, even if robot doesn't say so

        // for vision processing
        private ThunderVision.RGBValue[] _rgbValue = null;  // buffer for data going to ThunderVision
        private byte[] _bitmapRGB = null;                   // bitmap's RGB values for reading/writing
        private ThunderVision _thunderVision = new ThunderVision();
        private VisionDelegate _visionDelegate = null;
        private bool _visionProcessing = false;
	}
}
