﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using NetworkTables;
using NetworkTables.Tables;
using System.Windows.Forms;
using System.ComponentModel;
using System.Threading;

namespace ThunderDashboard
{
	static class DashboardNetworkTable
	{
		// delegate that gets called when table changes
		public delegate void DashboardNetworkTableDelegate(string key, string value);

        // network table string that we send when the table is first connected and disconnected
        public const string Connected = "$connected$";
        public const string Disconnected = "$disconnected$";

        // start a connection with the network table
        public static void Start(string ipAddress, DashboardNetworkTableDelegate tableDelegate, bool client = true)
        {
            // make sure nothing is running
            Disconnect();

            _ipAddress = ipAddress;
            _tableDelegate = tableDelegate;
            _client = client;

            // only create background worker if not already running
            if (_bw == null)
            {
                _bw = new BackgroundWorker();
                _bw.DoWork += new DoWorkEventHandler(DashboardNetworkTables_DoWork);
                _bw.RunWorkerAsync();
            }
        }

        public static void Restart(string ipAddress = null)
        {
            if (ipAddress != null)
            {
                _ipAddress = ipAddress;
            }
            _restart = true;
        }

        // stop connection with network table
        public static void Stop()
        {
            _shutdown = true;
        }

		// add/update a string in the network tables
		public static bool PutString(string key, string value)
		{
			if (_networkTable == null)
			{
				_lastError = "Table not initialized";
				return (false);
			}
			if (key == null)
			{
				_lastError = "Name was null";
				return (false);
			}

			try
			{
				_networkTable.PutString(key, ((value == null) ? "" : value));
				_lastError = "";
				return (true);
			}
			catch (Exception ex)
			{
				_lastError = "PutString Failed:\r\n" + ex.Message.ToString();
				return (false);
			}
		}

		// get a string from the network tables
		public static bool GetString(string key, ref string value, string defaultValue = "")
		{
			if (_networkTable == null)
			{
				_lastError = "Table not initialized";
				value = defaultValue;
				return (false);
			}
			if (key == null)
			{
				_lastError = "Name was null";
				value = defaultValue;
				return (false);
			}

			try
			{
				value = _networkTable.GetString(key, defaultValue);
				_lastError = "";
				return (true);
			}
			catch (Exception ex)
			{
				_lastError = "GetString Failed:\r\n" + ex.Message.ToString();
				value = defaultValue;
				return (false);
			}
		}

        // add/update a number in the network tables
        public static bool PutNumber(string key, decimal value)
        {
            try
            {
                return (PutNumber(key, Convert.ToDouble(value)));
            }
            catch
            {
                _lastError = "Cannot convert value";
                return (false);
            }
        }
        public static bool PutNumber(string key, double value)
        {
            if (_networkTable == null)
            {
                _lastError = "Table not initialized";
                return (false);
            }
            if (key == null)
            {
                _lastError = "Name was null";
                return (false);
            }

            try
            {
                _networkTable.PutNumber(key, value);
                _lastError = "";

                if (_tableListener != null)
                {
                    _tableListener.ValueChanged(null, key, _networkTable.GetNumber(key, value), NotifyFlags.NotifyLocal);
                }
                return (true);
            }
            catch (Exception ex)
            {
                _lastError = "PutSNumber Failed:\r\n" + ex.Message.ToString();
                return (false);
            }
        }

        // get a number from the network tables
        public static bool GetNumber(string key, ref double value, double defaultValue = 0)
        {
            if (_networkTable == null)
            {
                _lastError = "Table not initialized";
                value = defaultValue;
                return (false);
            }
            if (key == null)
            {
                _lastError = "Name was null";
                value = defaultValue;
                return (false);
            }

            try
            {
                value = _networkTable.GetNumber(key, defaultValue);
                _lastError = "";
                return (true);
            }
            catch (Exception ex)
            {
                _lastError = "GetNumber Failed:\r\n" + ex.Message.ToString();
                value = defaultValue;
                return (false);
            }
        }

        // get the last error the this class generated
        // typically call it when one of the other methods return false
        public static string LastError()
		{
			return (_lastError);
		}

        // get list of all key names from the network table
        // NOTE: this MUST be called from the UI thread and NOT
        //       directly from ValueChanged()
        private static object _lockUpdate = new object();
        public static bool ShowTableItem(ref ListView listView, string key, string value)
        {
            int idx;            // index of key within our list

            lock (_lockUpdate)
            {
                // make sure we are still connected
                if (!_connected)
                {
                    return (false);
                }

                // find the key in our list
                idx = _keyNames.IndexOf(key);
                if (idx < 0)
                {
                    string valueString = "";
                    double valueNumber = 0;

                    // key is not in our list, so re-get the entire list of keys
                    idx = UpdateEntireList(key);

                    listView.SuspendLayout();
                    listView.Items.Clear();
                    foreach (string keyName in _keyNames)
                    {
                        // if it is a section break, the use a blank;
                        // otherwise if it is string, then use GetString();
                        // otherwise use GetNumber()
                        if (string.IsNullOrEmpty(keyName))
                        {
                            valueString = "";
                        }
                        else if (_keysThatAreString.Contains(keyName))
                        {
                            DashboardNetworkTable.GetString(keyName, ref valueString, "$empty$");
                        }
                        else
                        {
                            DashboardNetworkTable.GetNumber(keyName, ref valueNumber, -99);
                            valueString = ValueToString(valueNumber);
                        }

                        // add row to the listview
                        ListViewItem lvi = new ListViewItem(new string[] { keyName, valueString });
                        listView.Items.Add(lvi);
                    }
                    listView.ResumeLayout();

                    // make sure we actually found it/ got table info
                    if (idx < 0)
                    {
                        return (false);
                    }
                }

                // update item's value
                listView.Items[idx].SubItems[1].Text = value;
                return (true);
            }
        }

        public static void EditItem(ref ListView listView, MouseEventArgs e)
        {
            ListViewHitTestInfo hti;
            string keyName = "";
            string valueString = "";
            bool isString = true;

            lock (_lockUpdate)
            {
                // make sure we are still connected
                if (!_connected)
                {
                    return;
                }

                // figure out which item it is and get its info
                try
                {
                    hti = listView.HitTest(e.X, e.Y);
                    if ((hti == null) || (hti.Item.SubItems.Count < 2))
                    {
                        return;
                    }
                    keyName = hti.Item.SubItems[0].Text;
                    valueString = hti.Item.SubItems[1].Text;
                    isString = _keysThatAreString.Contains(keyName);

                    // make sure it is something we want to let them edit
                    if (!keyName.EndsWith("_e"))
                    {
                        return;
                    }
                }
                catch
                {
                    return;
                }
            }

            // edit the value
            FormNetTblEdit form = new ThunderDashboard.FormNetTblEdit(keyName, valueString, isString);
            if (DialogResult.OK != form.ShowDialog())
            {
                return; // user cancelled
            }
            if (isString)
            {
                PutString(keyName, form.GetValueString);
            }
            else
            {
                PutNumber(keyName, form.GetValueNumber);
            }
        }

        // watches for the robot coming and going
        private static void DashboardNetworkTables_DoWork(object obj, DoWorkEventArgs eEventArgs)
        {
            bool robotThere;

            _restart = false;
            while (!_shutdown)
            {
                // see if the robot is even there (only if in client mode). if not, then wait 3 seconds and try again
                robotThere = !_client || Program.PingRobot(_ipAddress);

                // if already connected OR the robot is not there; then wait 3 seconds and check again
                if (_connected || !robotThere || _restart)
                {
                    // if robot is not there but we thought we were connected, then unload tables
                    if (!robotThere && _connected)
                    {
                        Disconnect();
                    }

                    // wait in 100ms intervals in case user closes the form while we were waiting
                    for (int i = 0; ((i < 30) && !_restart && !_shutdown); i++)
                    {
                        Thread.Sleep(100);
                    }

                    // if we need to shutdown, do that now
                    if (_shutdown)
                    {
                        Disconnect();
                        return;
                    }

                    // user wants to restart?
                    if (_restart)
                    {
                        _restart = false;
                        Disconnect();
                        continue;
                    }

                    // try again to see if robot is still there
                    continue;
                }

                // if here, then not connected but the robot is now there
                try
                {
                    // pick mode and initialize
                    if (_client)
                    {
                        NetworkTable.SetIPAddress(_ipAddress);
                        NetworkTable.SetClientMode();
                    }
                    else
                    {
                        NetworkTable.SetServerMode();
                    }
                    NetworkTable.Initialize();

                    // get reference to network table
                    _networkTable = NetworkTable.GetTable(_networkTableName);
                    if (_tableDelegate != null)
                    {
                        _tableListener = new DashboardNetworkTableListener(_tableDelegate);
                        _networkTable.AddTableListenerEx(_tableListener, (NotifyFlags.NotifyDelete
                                                                            | NotifyFlags.NotifyImmediate
                                                                            | NotifyFlags.NotifyLocal
                                                                            | NotifyFlags.NotifyNew
                                                                            | NotifyFlags.NotifyUpdate)); // we care about everything
                        _tableListener.ValueChanged(null, Connected, "", NotifyFlags.NotifyLocal);
                    }
                    _connected = true;
                }
                catch
                {
                    // cleanup, then try again
                    Disconnect();
                }
            }

            Disconnect();
        }

        // stop connection with network table
        private static void Disconnect()
        {
            lock (_lockUpdate)
            {
                try
                {
                    // only close table if we actually connected to it
                    if (_connected)
                    {
                        NetworkTable.Shutdown();
                    }
                }
                catch
                {
                }

                // let them know we disconnected
                if (_tableListener != null)
                {
                    _tableListener.ValueChanged(null, Disconnected, "", NotifyFlags.NotifyLocal);
                }

                _connected = false;
                _networkTable = null;
                _tableListener = null;
            }
        }

        // gets entire list of items from the network tabble
        private static int UpdateEntireList(string key)
        {
            _keyNames.Clear();

            if (_networkTable == null)
            {
                _lastError = "Table not initialized";
                return (-1);
            }
            if (key == null)
            {
                _lastError = "Key was null";
                return (-1);
            }

            // get all strings and numbers
            HashSet<string> keys = _networkTable.GetKeys(NetworkTables.Native.NtType.String);
            foreach (string keyName in keys)
            {
                _keyNames.Add(keyName);             // add to overall list
                _keysThatAreString.Add(keyName);    // remember it was a string
            }
            keys = _networkTable.GetKeys(NetworkTables.Native.NtType.Double);
            foreach (string keyName in keys)
            {
                _keyNames.Add(keyName);             // add to overall list
            }

            // sort the overall list
            _keyNames.Sort();

            // insert blanks between sections
            string sectionPrev = "";
            string sectionName;
            string[] astr;
            for (int i = 0; i < _keyNames.Count; i++)
            {
                // split off the first part to get section name (up to underscorce)
                sectionName = "";
                astr = _keyNames[i].Split(new char[] { '_' });
                if (astr.Length > 1)
                {
                    sectionName = astr[0];
                }

                // if first item in list, then it is the name of the first section
                if (i == 0)
                {
                    sectionPrev = sectionName;
                    continue;
                }

                // if new section, add a blank line
                if (sectionPrev != sectionName)
                {
                    sectionPrev = sectionName;
                    _keyNames.Insert(i, "");
                    i++;    // skip the entire this entry we just put in
                }
            }

            return (_keyNames.IndexOf(key));
        }

        // this is done so we do not get scientific notation
        private static string ValueToString(double value)
        {
            return (value.ToString("0." + new string('#', 339)));
        }

        // private class to handle getting updates when table is changed
        private class DashboardNetworkTableListener : ITableListener
		{
			public DashboardNetworkTableListener(DashboardNetworkTableDelegate tableDelegate)
			{
				_delegate = tableDelegate;
			}

			public void ValueChanged(ITable source, string key, object value, NotifyFlags flags)
			{
                if (_delegate != null)
                {
                    if (value is double)
                    {
                        _delegate(key, ValueToString((double)value));
                    }
                    else
                    {
                        _delegate(key, value.ToString());
                    }
				}
			}

			private DashboardNetworkTableDelegate _delegate = null;
		}

        private static string _ipAddress = null;
        private static DashboardNetworkTableDelegate _tableDelegate = null;
        private static bool _client = true;
        private static BackgroundWorker _bw = null;
        private static volatile bool _connected = false;
        private static volatile bool _restart = false;
        private static volatile bool _shutdown = false;

        // name of the table created by the robot
        private const string _networkTableName = "SmartDashboard";
		private static NetworkTable _networkTable = null;
		private static string _lastError = "";
		private static DashboardNetworkTableListener _tableListener = null;

        // current list of items in the network table that we care about
        private static List<string> _keyNames = new List<string>();
        private static List<string> _keysThatAreString = new List<string>();
    }
}
