﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ThunderDashboard
{
    public partial class FormMain
    {
        // start up all of the light processing
        private void LightsStart()
        {
            // clear designer text
            labelLightStatus.Text = "";
            labelLightPower.Text = "";

            // get the COM port list initialized.
            // the _bwComPorts backroundworker will fill in the ports, which
            // in turn will trigger things to read the light patterns
            UpdateComPortList();

            // check for COM ports in a thread. that way we can detect
            // when they come and go. but do not setup if already done
            if (_bwComPorts == null)
            {
                _bwComPorts = new BackgroundWorker();
                _bwComPorts.DoWork += _bwComPorts_DoWork;
                _bwComPorts.WorkerReportsProgress = true;
                _bwComPorts.ProgressChanged += _bwComPorts_ProgressChanged;
                _bwComPorts.RunWorkerCompleted += _bwComPorts_RunWorkerCompleted;
                _bwComPorts.RunWorkerAsync();
            }
        }

        // stop all of the lights processing
        private void LightsStop()
        {
            // if already stopped, then nothing to do
            if (_bwComPorts == null)
            {
                return;
            }

            // if busy, then tell it to stop
            if (_bwComPorts.IsBusy)
            {
                _bwComPorts.CancelAsync();
                return;
            }

            // shouldn't get here, but if we do, then clean up the object
            _bwComPorts = null;
        }

        // update the COM list on the UI (looks at _comPorts)
        private void UpdateComPortList()
        {
            int idx = 0;    // assume last selection not found

            // don't let two people play with the list at the same time
            // (i.e. background worker might be replacing the list again, or user selected another entry)
            lock (_lockComList)
            {
                // refill the list
                comboBoxComPort.SuspendLayout();                // don't draw anything until we are all done
                comboBoxComPort.Items.Clear();                  // remove all previous entries
                comboBoxComPort.Items.Add("none");              // be able to select no port
                for (int i = 0; i < _comPorts.Count; i++)       // fill with new list
                {
                    comboBoxComPort.Items.Add(_comPorts[i]);
                }
                comboBoxComPort.ResumeLayout();                 // all done so re-draw

                // find the user's desired selection
                for (int i = 0; i < comboBoxComPort.Items.Count; i++)
                {
                    if (comboBoxComPort.Items[i].ToString() == Properties.Settings.Default.LightsComPort)
                    {
                        idx = i;                                // found it
                        break;
                    }
                }
            }

            // now select the COM port
            // NOTE: do this outside the lock because this will trigger a
            //       selection change event, and that will want to do locking
            comboBoxComPort.SelectedIndex = idx;

            // if we lost our selection, then force things to be updated
            if ((idx <= 0) && !string.IsNullOrEmpty(_port) && (_port != _portNone))
            {
                comboBoxComPort_SelectedIndexChanged(null, null);
            }
        }

        // background worker thread that will watch for COM ports coming and going
        private void _bwComPorts_DoWork(object sender, DoWorkEventArgs e)
        {
            BackgroundWorker bw = sender as BackgroundWorker;
            string[] comPorts;
            bool updateList;

            while (true)
            {
                // get list of current COM ports
                comPorts = SerialPort.GetPortNames();
                updateList = false;     // assume list didn't change

                // see if the list changed
                if (comPorts.Length != _comPorts.Count)
                {
                    // we have a different number of com ports, so update the list
                    updateList = true;
                }
                else
                {
                    // the length is the same, so compare each string to see if there are any changes
                    for (int i = 0; i < comPorts.Length; i++)
                    {
                        if (comPorts[i] != _comPorts[i])
                        {
                            // name mistmatch, so the list changed
                            updateList = true;
                        }
                    }
                }

                // if we need to update the list, then copy new list over
                if (updateList)
                {
                    // do not update the list while the UI is looking at it
                    lock (_lockComList)
                    {
                        _comPorts.Clear();
                        for (int i = 0; i < comPorts.Length; i++)
                        {
                            _comPorts.Add(comPorts[i]);
                        }
                    }

                    // now that it is updated, tell UI to update itself
                    bw.ReportProgress(1);
                }

                // check to see if the power changed
                if (SystemInformation.PowerStatus.PowerLineStatus != _powerStatus)
                {
                    _powerStatus = SystemInformation.PowerStatus.PowerLineStatus;
                    bw.ReportProgress(2);
                }

                // wait before we bother to check for port changes (do not need to spam the PC)
                System.Threading.Thread.Sleep(3000);
            }
        }

        // called when background worker has something for the UI to do
        private void _bwComPorts_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            // what do we need to update?
            switch (e.ProgressPercentage)
            {
                default:
                    break;

                case 1:
                    // update the COM list
                    UpdateComPortList();
                    break;

                case 2:
                    // power changed so update lights
                    UpdateListboxLightPattern();
                    if ((_powerStatus == PowerLineStatus.Offline))
                    {
                        labelLightPower.Text = "running on battery";
                        LightPatternSet("0#");
                    }
                    else
                    {
                        labelLightPower.Text = "";
                        listBoxLightPattern_SelectedIndexChanged(null, null);
                    }
                    break;
            }
        }

        // called when the background thread is complete
        private void _bwComPorts_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            _bwComPorts = null;
        }

        // enable/disable listbox depending on whether we can send down light patterns or not
        private void UpdateListboxLightPattern()
        {
            // dsable iif we do not have light patterns OR we are running on battery
            bool enable = _haveLightPatterns && (_powerStatus != PowerLineStatus.Offline);

            labelLightPattern.Enabled = enable;
            listBoxLightPattern.Enabled = enable;
        }

        // called when the a COM port selection is made/changed
        private void comboBoxComPort_SelectedIndexChanged(object sender, EventArgs e)
        {
            try
            {
                // if something is not selected (or disabled), then no board to talk to
                if (comboBoxComPort.SelectedIndex < 0)
                {
                    _port = "";                 // indicate we have no board

                    // gray controls so user knows they are not connected
                    _haveLightPatterns = false;
                    UpdateListboxLightPattern();
                    listBoxLightPattern.Items.Clear();
                    return;
                }

                // get the newly selected port
                _port = comboBoxComPort.SelectedItem.ToString();

                // get list from board, if not already getting it
                // NOTE: we do this in a background worker because
                //       opening the serial port can take a long time
                if (_bwLightPatterns == null)
                {
                    _haveLightPatterns = false;
                    UpdateListboxLightPattern();
                    labelLightStatus.Text = "reading light patterns, please wait...";
                    labelComPort.Enabled = false;
                    comboBoxComPort.Enabled = false;
                    buttonReset.Enabled = false;
                    _bwLightPatterns = new BackgroundWorker();
                    _bwLightPatterns.DoWork += _bwLightPatterns_DoWork;
                    _bwLightPatterns.RunWorkerCompleted += _bwLightPatterns_RunWorkerCompleted;
                    _bwLightPatterns.RunWorkerAsync();
                }

                // save their selection, but after we get the list, in case the save fails
                // if not 'none', then we always save the value.
                // if it is 'none', then we only want to save it if we have
                //    more than 1 item in list. because that means the user
                //    had a choice and specifically said 'none'. but if the
                //    list has just 'none' in it, then most likely the board
                //    is not plugged in yet; in which case we do not want to
                //    lose their selection. so when the board is plugged in
                //    later we will pick their port from before.
                // since in both cases the list has more than 1 item, we can
                // always just do that check.
                if (comboBoxComPort.Items.Count > 1)
                {
                    Properties.Settings.Default.LightsComPort = _port;
                    Properties.Settings.Default.Save();
                }
            }
            catch
            {
            }
        }

        // result types when trying to read the patterns
        private enum LightResult
        {
            Success,
            NoPort,
            Timeout,
            RetryLimit,
        };

        // background worker that gets the list of patterns from the board
        private void _bwLightPatterns_DoWork(object sender, DoWorkEventArgs e)
        {
            SerialPort serialPort = null;
            string line;

            // clear out the current list
            _lightPatterns = new List<string>();

            // if no port, then nothing to do, so just return
            if (string.IsNullOrEmpty(_port) || (_port == _portNone))
            {
                e.Result = LightResult.NoPort;
                return;
            }

            // retry until we get a list of patterns (sometimes the board doesn't do it the first time)
            for (int i = 0; i < 3; i++)
            {
                try
                {
                    // don't let two people play with the list at the same time
                    lock (_lockSerialPort)
                    {
                        // clear out the current list
                        _lightPatterns = new List<string>();

                        // open the new port
                        serialPort = new SerialPort(_port, 115200, Parity.None, 8, StopBits.One);
                        serialPort.Open();
                        serialPort.DiscardInBuffer();       // clear out any leftover data
                        serialPort.ReadTimeout = 2000;      // do not wait forever (in case board is mis behaving)
                        serialPort.Write("?");              // ask board for the list
                        while (true)                        // read all lines from the board
                        {
                            line = serialPort.ReadLine();   // each line is a separate choice/selection
                            if (string.IsNullOrEmpty(line))
                            {
                                break;                      // hit the end of the list, so we are done
                            }
                            _lightPatterns.Add(line);       // save the choice
                        }
                        serialPort.Close();
                    }
                }
                catch
                {
                    // could get here because the read timed out, or error opening the COM port
                    if ((serialPort != null) && serialPort.IsOpen)
                    {
                        serialPort.Close();
                    }
                    e.Result = LightResult.Timeout;  // timeout
                }

                // if we got at least one pattern then we are done
                if (_lightPatterns.Count > 1)
                {
                    e.Result = LightResult.Success;
                    return;
                }
            }

            e.Result = LightResult.RetryLimit; // too many retries
        }

        // called when the background worker is done getting the list of patterns
        private void _bwLightPatterns_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            // re-fill the list
            listBoxLightPattern.SuspendLayout();
            listBoxLightPattern.Items.Clear();
            foreach (string pattern in _lightPatterns)
            {
                listBoxLightPattern.Items.Add(pattern);
            }
            if (listBoxLightPattern.Items.Count > 0)
            {
                // ungray controls so user knows they are connected
                _haveLightPatterns = true;
                UpdateListboxLightPattern();
            }
            listBoxLightPattern.ResumeLayout();

            // now find the last item the user wanted
            for (int i = 0; i < listBoxLightPattern.Items.Count; i++)
            {
                if (listBoxLightPattern.Items[i].ToString().StartsWith(Properties.Settings.Default.LightPattern))
                {
                    listBoxLightPattern.SelectedIndex = i;
                    break;
                }                    
            }
            if ((listBoxLightPattern.SelectedIndex < 0) && (listBoxLightPattern.Items.Count > 0))
            {
                // make sure something is selected
                listBoxLightPattern.SelectedIndex = 0;
            }

            // let us know we are now done with getting the pattern's from the board
            switch ((LightResult)e.Result)
            {
                default:
                    labelLightStatus.Text = "Unexpected error: " + (LightResult)e.Result;
                    break;

                case LightResult.NoPort:
                    labelLightStatus.Text = "no COM Port selected";
                    break;
                case LightResult.RetryLimit:
                    labelLightStatus.Text = "too many retries";
                    break;
                case LightResult.Success:
                    labelLightStatus.Text = "";
                    break;
                case LightResult.Timeout:
                    labelLightStatus.Text = "light board didn't respond in time";
                    break;

            }
            labelComPort.Enabled = true;
            comboBoxComPort.Enabled = true;
            buttonReset.Enabled = true;
            _bwLightPatterns = null;
        }

        // called when a light pattern selection has been made/changed
        private void listBoxLightPattern_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (listBoxLightPattern.SelectedIndex >= 0)
            {
                LightPatternSet(listBoxLightPattern.SelectedItem.ToString());
            }
        }

        // send the light pattern to the board
        private void LightPatternSet(string pattern)
        {
            SerialPort serialPort = null;

            // if no port, then nothing to do, so just return
            if (string.IsNullOrEmpty(_port) || (_port == _portNone))
            {
                return;
            }

            try
            {
                // don't let two people play with the list at the same time
                lock (_lockSerialPort)
                {
                    string[] values = pattern.Split(new string[] { " - " }, StringSplitOptions.RemoveEmptyEntries);
                    serialPort = new SerialPort(_port, 115200, Parity.None, 8, StopBits.One);
                    serialPort.Open();
                    serialPort.Write(values[0]);
                    serialPort.Close();

                    // save the pattern. do it after we set the pattern in case the save fails
                    Properties.Settings.Default.LightPattern = values[0];
                    Properties.Settings.Default.Save();
                }
            }
            catch
            {
                if ((serialPort != null) && serialPort.IsOpen)
                {
                    serialPort.Close();
                }
            }
        }

        // user wants us to re-read the light patterns
        private void buttonReset_Click(object sender, EventArgs e)
        {
            comboBoxComPort_SelectedIndexChanged(null, null);
        }

        private BackgroundWorker _bwComPorts = null;                // tracks COM port changes
        private List<string> _comPorts = new List<string>();        // available COM ports on the PC
        private object _lockComList = new object();                 // prevent two threads from accessing _comPorts
        private volatile string _port = "";                         // currently selected COM port
        private const string _portNone = "none";                    // text for selecting No port

        private object _lockSerialPort = new object();              // prevent two threads from accessing the board

        private BackgroundWorker _bwLightPatterns = null;           // gets light patterns
        private List<string> _lightPatterns = new List<string>();   // available COM ports on the PC

        private bool _haveLightPatterns = false;                                 // COM port connected and have light patterns
        private volatile PowerLineStatus _powerStatus = PowerLineStatus.Unknown; // running on battery or not
    }
}