﻿/////////////////////////////////////////////////////////////////////////////////////////
//
// Uses network table entries to drive chart.
//
// All entries must be a comma separated string.
// The first item in the string corresponds to the time (i.e. X axis).
// The rest of the items are for eaach line of the chart.
// So if you want 3 lines, then there needs to be 4 items in the string (e.g. "0.1,3.4,5,6").
//
// Network table names must end with the following:
//      _plotmin                the minimum values for the Y axis
//      _plotmax                the maximum values for the Y axis
//                                  this is ignored for time, but must put something there (i.e. 0)
//      _plotlab                optional labels for the lines (used for legend)
//      _plotctr                optionally add a center-line (0 means no line, 1 means line)
//      _plot                   the data values for the lines
//
// The number of items in all of these must match the first data value of the chart.
// If the min & max are not supplied (or not the correct number of items), then the data
// is used to figured it out.
//
// The line colors are determine by this code (see _chartColors).
//
// The _plot data is not saved until AFTER the 'Start' button is clicked on the Chart
// tab. The other _plot* values can come at anytime.
//
// The main UI thread will take the _plot value from the robot and store it (i.e. _chartValues).
// The background waits for new values to be stored, then converts them to what the chart
// needs (i.e. splits the comma separated values into doubles). It then calls back into the
// main UI thread to re-draw the chart. The background thread sleeps (100ms originally), so
// tweak ChartBackgroundThread() for longer sleep if chart drawing takes too long.
//
/////////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Net.NetworkInformation;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using OxyPlot;
using OxyPlot.Axes;
using OxyPlot.Series;
using System.ComponentModel;
using System.Threading;

namespace ThunderDashboard
{
    public partial class FormMain
    {
        // initialize chart controls
        private void ChartInit()
        {
            numericUpDownChart.Value = _chartTimeToView = Properties.Settings.Default.ChartTime;
            checkBoxChartLive.Checked = _chartUpdateLive = Properties.Settings.Default.ChartLive;
            checkBoxChartAutoScroll.Checked = _chartAutoScroll = Properties.Settings.Default.ChartAutoScroll;
        }

        // get the network table values from the robot
        private bool ChartValue(string key, string value)
        {
            if (key.EndsWith("_plotmin"))       // minimum values
            {
                _plotMinValues = value;
                ChartFill();
                return (true);
            }

            if (key.EndsWith("_plotmax"))       // maximum values
            {
                _plotMaxValues = value;
                ChartFill();
                return (true);
            }

            if (key.EndsWith("_plotlab"))       // labels
            {
                _plotLabs = value;
                ChartFill();
                return (true);
            }

            if (key.EndsWith("_plotctr"))       // center lines?
            {
                _plotCtrs = value;
                ChartFill();
                return (true);
            }

            if (key.EndsWith("_plot"))          // the data
            {
                // only save the data if 'Start' was pressed
                if (_plotSaveValues)
                {
                    lock (_chartLock)                // make sure background thread isn't looking at it
                    {
                        _chartValues.Add(value);
                    }
                }
                return (true);
            }
            return (false);
        }

        // selection for how many seconds to show on the screen has changed
        private void numericUpDownChart_ValueChanged(object sender, EventArgs e)
        {
            // always save the user's selection
            try
            {
                _chartTimeToView = Convert.ToInt32(numericUpDownChart.Value);
                Properties.Settings.Default.ChartTime = _chartTimeToView;
                Properties.Settings.Default.Save();
            }
            catch
            {
            }

            // only update chart if we are not in the middle of gathering values
            if (!_plotSaveValues)
            {
                ChartFill();
            }
        }

        // 'Update live' checkbox changed
        private void checkBoxChartLive_CheckedChanged(object sender, EventArgs e)
        {
            _chartUpdateLive = checkBoxChartLive.Checked;
            try
            {
                Properties.Settings.Default.ChartLive = _chartUpdateLive;
                Properties.Settings.Default.Save();
            }
            catch
            {
            }
        }

        // 'Autoscroll; checkbox was changed
        private void checkBoxChartAutoScroll_CheckedChanged(object sender, EventArgs e)
        {
            _chartAutoScroll = checkBoxChartAutoScroll.Checked;
            try
            {
                Properties.Settings.Default.ChartAutoScroll = _chartAutoScroll;
                Properties.Settings.Default.Save();
            }
            catch
            {
            }
        }

        // 'Start' button was pressed
        private void buttonChartStart_Click(object sender, EventArgs e)
        {
            _plotSaveValues = true;                     // we can now save any incoming values
            _chartShutdown = false;                     // clear flag so the thread doesn't immediately stop

            buttonChartStop.Enabled = true;             // allow user to press Stop
            buttonChartStop.Focus();                    // give Stop focus before we take away Start
            buttonChartStart.Enabled = false;           // do not allow them to press Start again

            // starting over, so clear everything; but make sure background thread isn't looking at
            lock (_chartLock)
            {
                _chartValues = new List<string>();      // clear values from robot
                _chartSeries = new List<LineSeries>();  // clear our series data
                _chartValuesRead = 0;                   // we have not copied any chart data yet
            }
            ChartFill();                                // clear chart now that values are cleared

            // only create background worker if not already running
            if (_chartBw == null)
            {
                _chartBw = new BackgroundWorker();
                _chartBw.WorkerReportsProgress = true;
                _chartBw.DoWork += new DoWorkEventHandler(ChartBackgroundThread);
                _chartBw.RunWorkerCompleted += ChartRunWorkerCompleted;
                _chartBw.ProgressChanged += ChartProgressChanged;
            }

            // re-start the background thread
            _chartBw.RunWorkerAsync();
        }

        // 'Stop' was pressed
        private void buttonChartStop_Click(object sender, EventArgs e)
        {
            _plotSaveValues = false;                    // stop saving any incoming data

            // tell the background thread to stop. ChartRunWorkerCompleted() will be
            // called after the thread has stopped
            _chartShutdown = true;
        }

        // panel that holds chart changed, so re-draw it to fit
        private void panelChart_SizeChanged(object sender, EventArgs e)
        {
            ChartFill();
        }

        // save entire chart
        private void buttonChartSaveAll_Click(object sender, EventArgs e)
        {
            ChartSave(true);
        }

        // save only the part of the chart that is being shown
        private void buttonChartSaveScreen_Click(object sender, EventArgs e)
        {
            ChartSave(false);
        }

        // do the actual save
        private void ChartSave(bool saveFull)
        {
            // make sure there is a chart to save
            if (plotChart.Model == null)
            {
                MessageBox.Show("There is no plot to save");
                return;
            }

            try
            {
                // prompt for the file name
                SaveFileDialog fileDialog = new SaveFileDialog();
                fileDialog.Filter = "Image Files|*.png";
                fileDialog.Title = "Select a file";
                fileDialog.CheckFileExists = false;
                fileDialog.CheckPathExists = true;
                fileDialog.DefaultExt = ".png";
                if (fileDialog.ShowDialog() != DialogResult.OK)
                {
                    return;
                }

                // save file name, then force dialog to close; and wait for it to close
                string filename = fileDialog.FileName;
                fileDialog = null;

                // save all or just what is shown
                if (saveFull)
                {
                    OxyPlot.WindowsForms.PngExporter.Export(plotChart.Model, filename, plotChart.Width, plotChart.Height);
                }
                else
                {
                    // wait for the dialog to not be visible (so it isn't part of the image)
                    System.Threading.Thread.Sleep(500);

                    using (Bitmap bmp = new Bitmap(this.Width, this.Height, System.Drawing.Imaging.PixelFormat.Format32bppArgb))
                    {
                        Graphics graphics = Graphics.FromImage(bmp);
                        graphics.CopyFromScreen(this.Left, this.Top, 0, 0, new Size(this.Width, this.Height), CopyPixelOperation.SourceCopy);
                        bmp.Save(filename, System.Drawing.Imaging.ImageFormat.Png);
                    }
                }
            }
            catch
            {
                MessageBox.Show("Unexpected error trying to save the file");
            }
        }

        // background thread is complete (would be called after 'Stop' was pressed
        private void ChartRunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            buttonChartStart.Enabled = true;            // allow user to press Start
            buttonChartStart.Focus();                   // give Start focus before we take away Stop
            buttonChartStop.Enabled = false;            // do not allow them to press Stop again

            // make sure chart data is fully up to date.
            // if 'Update live' is off, then we still need to re-draw because while we have data, we didn't draw it
            if (ChartCheckForUpdate() || !_chartUpdateLive)
            {
                ChartFill();
            }
        }

        // background worker thread
        private void ChartBackgroundThread(object obj, DoWorkEventArgs eEventArgs)
        {
            bool update;

            do
            {
                // give the main thread a chance to get some values AND for last chart to be drawn
                Thread.Sleep(100);

                // see if there are any new values
                // (keep array locked for as little time as possible)
                lock (_chartLock)
                {
                    update = ChartCheckForUpdate();
                }

                // if we updated things, then ask main thread to re-draw the graph
                if (update && _chartUpdateLive)
                {
                    (obj as BackgroundWorker).ReportProgress(0);
                }
            } while (!_chartShutdown);          // continue until we are asked to stop
        }

        // background is asking main thread to update the chart
        private void ChartProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            ChartFill();
        }

        // check to see if any new values have come in from the robot
        private bool ChartCheckForUpdate()
        {
            bool update = false;

            if (_chartValues.Count != _chartValuesRead)
            {
                update = true;                          // more values came in
                ChartAddValues(_chartValuesRead);       // add all the new values to the chart lines
                _chartValuesRead = _chartValues.Count;  // remember what we have just done
            }
            return (update);
        }

        // add data values that came from the robot and prep them for the chart
        private void ChartAddValues(int startingIndex)
        {
            List<double> valuesDouble;

            // if we do not have any series yet, then this is the first plot value so
            // we will use it to determine the number of series we need
            int plotSeries = _chartSeries.Count;
            if (plotSeries < 1)
            {
                // make sure we have at least one value
                if (_chartValues.Count < 1)
                {
                    return;
                }

                // convert all the list of strings into doubles to see how many series we need
                valuesDouble = ChartConvertValues(_chartValues[0], 0);
                if ((valuesDouble == null) || (valuesDouble.Count < 2)) // 1st num is X, so need at least 2
                {
                    return; // either we had a bad value or we didn't the correct number of values
                }

                // make a slot for each series
                while (_chartSeries.Count < (valuesDouble.Count - 1))
                {
                    _chartSeries.Add(new LineSeries());
                }
            }

            // make sure we are within the area bounds
            if ((startingIndex + 1) > _chartValues.Count)
            {
                return;
            }

            // convert all the values to plot data
            for (int ii = startingIndex; ii < _chartValues.Count; ii++)
            {
                valuesDouble = ChartConvertValues(_chartValues[ii], (_chartSeries.Count + 1));
                if (valuesDouble == null)
                {
                    continue;
                }

                // save all the points (start at 1 because the X axis value is in the first slot)
                for (int i = 1; i < valuesDouble.Count; i++)
                {
                    _chartSeries[i - 1].Points.Add(new DataPoint(valuesDouble[0], valuesDouble[i]));
                }
            }
        }

        // convert one comman separated robot value into a list of doubles for the chart
        private List<double> ChartConvertValues(string values, int requiredValues)
        {
            string[] valuesString = null;
            List<double> valuesDouble = new List<double>();

            // validate
            if (values == null)
            {
                return (null);
            }

            // split string to get value
            valuesString = values.Split(_separator);

            // if we do not have as many values as we require, then just return
            if ((requiredValues > 0) && (valuesString.Length != requiredValues))
            {
                return (null);
            }

            // convert all numbers to a value. we do all the conversion before do any point
            // in case one of the values throws an exception
            try
            {
                // convert first
                for (int i = 0; i < valuesString.Length; i++)
                {
                    valuesDouble.Add(Convert.ToDouble(valuesString[i]));
                }
                return (valuesDouble);
            }
            catch
            {
            }
            return (null);
        }

        // draw the chart based on the data that we have
        private void ChartFill()
        {
            List<double> plotMins = null;                       // minimum values converted to doubles
            List<double> plotMaxs = null;                       // maximum values converted to doubles
            string[] plotLabs = null;                           // labels
            string[] plotCtrs = null;                           // draw center line?
            List<LinearAxis> axisList = new List<LinearAxis>(); // each axis for the chart
            int axisMarginLeft = 0;                             // how much space we need for all the Y axis labels
            int axisMarginBotttom = 0;                          // space below X axis for labels
            int plotNumSeries = _chartSeries.Count;             // number of lines for the chart
            LinearAxis axis;

            // wipe out any previous info
            if (plotChart.Model != null)
            {
                plotChart.Model.Series.Clear();                 // if do not do this an expection will be thrown becuase we are re-using the array
                plotChart.Model.Axes.Clear();                   // assumig we would have the same problem here
                plotChart.Model = null;
            }

            // start with a new model
            PlotModel model = new PlotModel();

            // get the min and max values. we need to have both or none
            plotMins = ChartConvertValues(_plotMinValues, 0);// (plotNumSeries + 1));
            if (plotMins != null)
            {
                // have min values, so check for max values
                plotMaxs = ChartConvertValues(_plotMaxValues, 0);// (plotNumSeries + 1));
                if (plotMaxs == null)
                {
                    // we didn't have any max values so clear out the min values
                    plotMins = null;
                }
                else if (  (plotMins.Count != plotMaxs.Count)                           // # values do not match
                        || (plotMins.Count == 0)                                        // no min/max values at all
                        || ((plotNumSeries > 0) && (plotMins.Count != (plotNumSeries + 1)))   // have data but # min/max values does not match
                        )
                {
                    // we had max values, but counts didn't match up so we cannot use them
                    plotMins = plotMaxs = null; 
                }
                else if ((plotNumSeries == 0) && (plotMins.Count > 0))
                {
                    // we had min and max, but no data. so use the min/max value
                    // count to determine the number of series to show. this lets
                    // us show a blank graph with axis
                    plotNumSeries = plotMins.Count - 1;
                }
            }

            // make sure we have at least one series; otherwise we are done
            if (plotNumSeries < 1)
            {
                return;
            }

            // get the labels. there must be as many of them as values
            if (!string.IsNullOrEmpty(_plotLabs))
            {
                plotLabs = _plotLabs.Split(_separator);
                if (plotLabs.Length != (plotNumSeries + 1))
                {
                    plotLabs = null;
                }
            }

            // get the center line. ther must be as many of them as values
            if (!string.IsNullOrEmpty(_plotCtrs))
            {
                plotCtrs = _plotCtrs.Split(_separator);
                if (plotCtrs.Length != (plotNumSeries + 1))
                {
                    plotCtrs = null;
                }
            }

            // create time axis (let it figure out the min & max)
            axis = new LinearAxis();
            axis.Key = "X";                             // something so we know what it is
            axis.Position = AxisPosition.Bottom;        // this is the X axis
            if ((_chartSeries.Count < 1) || (_chartSeries[0].Points.Count < 1))
            {
                axis.Minimum = 0;
                axis.Maximum = _chartTimeToView;
            }
            else
            {
                axis.Minimum = _chartSeries[0].Points[0].X;
                axis.Maximum = _chartSeries[0].Points[_chartSeries[0].Points.Count - 1].X;
                if (axis.Maximum < _chartTimeToView)
                {
                    axis.Maximum = _chartTimeToView;
                }
            }
            if (plotLabs != null)
            {
                axisMarginBotttom += 20;
                axis.Title = plotLabs[0];
            }
            model.Axes.Add(axis);

            // start at 1, because the first value is the time (i.e. Y axis)
            for (int i = 1; i <= plotNumSeries; i++)
            {
                axis = new LinearAxis();

                axis.Key = i.ToString();                // just something unique, but something we know
                axis.Position = AxisPosition.Left;      // this is a Y axis
                axis.PositionTier = (plotNumSeries - i);// need to separate each axis so they do not end up on to of each other
                axis.AxisDistance = 10;                 // separate the axis labels from each other
                axis.TextColor = _chartColors[(i - 1) % _chartColors.Length]; // do mod incase we have more series then defined colors
                axis.TicklineColor = axis.TextColor;    // make everything the same color
                axis.TitleColor = axis.TextColor;       // make everything the same color

                // only set min/max if we have both values
                if ((plotMins != null) && (plotMaxs != null))
                {
                    axis.Minimum = plotMins[i];
                    axis.Maximum = plotMaxs[i];

                    if ((plotCtrs != null) && (plotCtrs[i] == "1"))
                    {
                        axis.ExtraGridlines = new Double[] { 0 };// ((axis.Minimum + axis.Maximum) / 2.0) };
                        axis.ExtraGridlineColor = axis.TextColor;
                    }
                }

                axisMarginLeft += 30 + (int)axis.AxisDistance;// make sure we have room for the axis on side of the graph

                model.Axes.Add(axis);
                axisList.Add(axis);
            }

            // give room for all axis
            // NOTE: needed a large top & bottom so the Y axis for -5 to 5 didn't get cut off
            model.PlotMargins = new OxyThickness(axisMarginLeft, 10, 10, axisMarginBotttom+30);

            // add all the data
            for (int i = 0; i < _chartSeries.Count; i++)
            {
                // map this to the proper axis (and use the same color)
                _chartSeries[i].XAxisKey = "X";
                _chartSeries[i].YAxisKey = (i + 1).ToString();
                _chartSeries[i].Color = _chartColors[i % _chartColors.Length]; // do mod incase we have more series then defined colors
                if (plotLabs != null) 
                {
                    _chartSeries[i].Title = plotLabs[i + 1];    // we have a label so use it
                }
                model.Series.Add(_chartSeries[i]);
            }

            // get the highest time value and make sure it isn't smaller than what the user wants to see
            double maxTime = _chartTimeToView;
            if ((_chartSeries.Count > 0) && (_chartSeries[0].Points.Count > 0))
            {
                maxTime = _chartSeries[0].Points[_chartSeries[0].Points.Count - 1].X;
                if (maxTime < _chartTimeToView)
                {
                    maxTime = _chartTimeToView;
                }
            }

            plotChart.Model = model;
            plotChart.Width = (int)((double)maxTime * ((double)(panelChart.Width - 2) / (double)_chartTimeToView));//-2 is to keep scrollbar from appearing when data less than _chartTimeToView
            plotChart.Height = panelChart.Height - 24;// leave room for scrollbar
            plotChart.Model.ResetAllAxes();

            // scroll to end of chart
            if (_chartAutoScroll)
            {
                using (Control c = new Control() { Parent = panelChart, Height = 1, Width = 1, Dock = DockStyle.Right })
                {
                    panelChart.ScrollControlIntoView(c);
                    c.Parent = null;
                }
            }
        }

        private int _chartTimeToView = 10;              // number of seconds to show in the viewer
        private volatile bool _chartUpdateLive = true;  // update chart as data comes in?
        private volatile bool _chartAutoScroll = true;  // scroll to end of chart as data comes in?

        private bool _plotSaveValues = false;           // are we saving the data? (i.e. was 'Start' button pressed)
#if debug
        private string _plotMinValues = "0,0,-1,-5";    // comma separated list of minimum values for each line
        private string _plotMaxValues = "0,10,1,5";     // comma separated list of maximum values for each line
        private string _plotLabs = "t,x,s,v";           // comma separated list of labels
        private string _plotCtrs = "0,0,1,0";           // comma separated 0/1 for whether to also draw a center line or not
#else
        private string _plotMinValues = "";             // comma separated list of minimum values for each line
        private string _plotMaxValues = "";             // comma separated list of maximum values for each line
        private string _plotLabs = "";                  // comma separated list of labels
        private string _plotCtrs = "";                  // comma separated 0/1 for whether to also draw a center line or not
#endif

        private BackgroundWorker _chartBw;              // background worker object

        // these are shared with the background thread and main UI thread.
        // the _chartValues is filled in as soon as it comes from the robot. then the background thread converts
        // them into the line data and calls back into the UI thread to redraw the chart
        private volatile bool _chartShutdown = false;                           // want background thread to shutdown?
        private volatile List<string> _chartValues = new List<string>();        // list of all data values from the robot
        private volatile List<LineSeries> _chartSeries = new List<LineSeries>();// all the line data, ready for the chart
        private volatile int _chartValuesRead = 0;                              // number of _chartValues the background thread has already converted

        private object _chartLock = new object();    // lock used around items that the background worker and main thread need to share

        // list of colors for the lines (code will wrap and re-use colors if we have more lines that colors)
        private static OxyColor[] _chartColors = { OxyColors.Red, OxyColors.Green, OxyColors.Blue, OxyColors.Black, OxyColors.Purple, OxyColors.PowderBlue };

        // how strings from robot are separated
        private static char[] _separator = new char[] { ',' };
    }
}
