﻿// For a new year:
//
// 1. copy the form folder (e.g. Form2018)
//       - right-click the folder (e.g. Form2018)
//       - select 'Copy'
//       - right-click the ThunderDahsboard project (looks like a folder)
//       - select 'Paste'
// 2. rename the copied folder to the current year
//       - right-click the folder (e.g. Form2018)
//       - select 'Rename'
//       - type in the year (e.g. 2019)
// 3. rename the form to current year
//       - right-click file (e.g. Form2018.cs)
//       - select 'Rename'
//       - update the year (e.g. Form2019.cs)
// 4. update the form's class and constructor
//       - right-click file (e.g. Form2019.cs)
//       - select 'View Code'
//       - update the class name (e.g. change 'public partial class Form2018: FormBase' to 'public partial class Form2019: FormBase')
//       - update the constructor name (e.g. change 'public Form2018()' to 'public Form2019()')
// 5. update form's designer file
//       - expand .cs file (click arrow to left of Form2019.cs)
//       - right-click file (e.g. Form2019.Designer.cs)
//       - select 'View Code'
//       - update the class name (e.g. change 'partial class Form2018' to 'partial class Form2019')
//       - partial class Form2017
// 6. add year to string  list at top of FormMain() in FormMain.cs
// 7. update the form's constructor to match what is desired (i.e. Form2019())
// 8. update form, in Designer, with desired controls
// 9. update NetworkTableUpdated() in the form's .cs file for the year's specific network table entries
// 10. update control pictures (easier to just update the files in Windows File explorer)
// 11. change the default camera image
//       - open the form in the designer
//       - select the camera object
//       - select the 'Image' property
//       - click the button with the 3 dots at appears
//       - select 'Lcoal resource'
//       - click 'Import...'
//       - find the Camera.png in the proper folder
//       - click Open
//       - click OK
//
// Network Table items:
//    Common names (be sure to re-use these):
//      Auto_Mode                       dashboard sets this to the user selected auto mode (this is a number)
//      thunderdashboard_drivereverse   robot is driving in reverse: 1 = yes
//      thunderdashboard_frontcamera    camera feed is the front camera: 1 = yes
//      vision_Start                    robot tells dashboard to start/stop vision procesing: true/false
//      vision_Counter                  gets incremented (by the dashboard) each time an image is processed, starting at 1
//
//      thunderdashboard_inpitmode      robot is in pit mode: 1 = yes  (display background will be an 'odd' color)
//      thunderdashboard_gyro           robot's gyro is being calirbrated: 1 = yes (display background will be red)
//      thunderdashboard_yellow         robot wants display background to be yellow: 1 = yes
//      thunderdashboard_green          robot wants display background to be green: 1 = yes
//      thunderdashboard_haveball       does same as thunderdashboard_green 


using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ThunderDashboard
{
    public partial class FormMain : Form
    {
        // creates the application screens
        public FormMain()
        {
            // ***1511***
            // add the new year to this list; MUST be the first item in the list
            string[] robotYears = new string[] { "2018", "2017", "2016", "2015" };


            // ***********************************************
            // leave the rest of method alone

            // check the command line
            // see if we were started and asked to run as fake robot (i.e. be the Network Table Server)
            string[] args = Environment.GetCommandLineArgs();
            if ((args.Length == 2) && (args[1] == "-RunAsServer"))
            {
                _runAsServer = true;
            }

            // initialize the common controls
            InitializeComponent();

            // load all the settings from the last time dashboard was run; and validate them
            Properties.Settings.Default.Reload();
            string year = Properties.Settings.Default.Year;
            if (string.IsNullOrEmpty(year))
            {
                year = robotYears[0];
            }
            string teamNumber = Properties.Settings.Default.Team;
            if (string.IsNullOrEmpty(teamNumber))
            {
                teamNumber = "1511";
            }

            // set the dashboard's title
            Text = "Thunder Dashboard " + year + ": " + teamNumber;

            // fill robot list with the supported robot years and initialize Team & Year
            foreach (string robotYear in robotYears)
            {
                comboBoxYear.Items.Add(robotYear);
            }
            comboBoxYear.SelectedItem = year;
            textBoxTeam.Text = teamNumber;

            // get the form for given year
            _formBase = (FormBase)System.Reflection.Assembly.GetExecutingAssembly().CreateInstance("ThunderDashboard.Form" + year);
            if (_formBase == null)
            {
                // programmer error; form not found
                MessageBox.Show("Could not find the form for year: " + year);
                tabControlDashboard.TabPages.Clear(); // make sure we remove anything from the designer
                tabControlDashboard.TabPages.Add(tabPageSettings);
                return;
            }

            // convert the form to a tab page
            _formBase.TopLevel = false;
            _formBase.Visible = true;
            TabPage pageForm = new TabPage();
            pageForm.SizeChanged += PageForm_SizeChanged; // need to know when tab changes size, so we can scale the form
            pageForm.Text = _formBase.Text;
            pageForm.Controls.Add(_formBase);

            _formBase.TeamNumber = teamNumber; // tell the form what the team number is

            // add the forms
            tabControlDashboard.TabPages.Clear(); // make sure we remove anything from the designer
            tabControlDashboard.TabPages.Add(pageForm);
            if (_formBase.TabLightsPsoc42xx)
            {
                tabControlDashboard.TabPages.Add(tabPageLightsPsoc42xx);
                if (!_runAsServer) { LightsStart(); } // since we have light, start is background processor
            }
            tabControlDashboard.TabPages.Add(tabPageSettings);
            if (_formBase.TabChart)
            {
                tabControlDashboard.TabPages.Add(tabPageChart);
                if (!_runAsServer) { ChartInit(); } // since we have charts, initialize it
            }
            tabControlDashboard.TabPages.Add(tabPageRoboDebug);
            tabPageRoboDebug.Text = _formBase.DebugTitle;

            // show this years graphics
            CheckControllerPictures(_formBase.ControllerDriver, _formBase.ControllerAux);

            // if we are suppose to run as the network table server, then do that now
            if (_runAsServer)
            {
                buttonServer_Click(null, null);
                return;
            }

            // connect to network table start up common items for the base form
            DashboardNetworkTable.Start(_formBase.RobotIpAddress, NetworkTableDelegate);
            _formBase.Start();
        }

        // called when the main window is first shown
        private void Form_Shown(object sender, EventArgs e)
        {
            // if we are running in server mode, then leave the form as is
            if (_runAsServer)
            {
                return;
            }

            // put window in upper-left corner of screen
            Location = new Point(0, 0);
            WindowState = FormWindowState.Normal;   // Driver's Station starts us minimized, so override that

            // sync colors
            SetBackColors();
        }

        // called when the main window is being closed
        private void Form_FormClosed(object sender, FormClosedEventArgs e)
        {
            DashboardNetworkTable.Stop();
            if (_formBase != null)
            {
                _formBase.Stop();
            }

            // see if we started a server (i.e. fake robot); close it if we did
            NetworkTableServerStop();
        }

        // called with the tab page, that the robot's form is on, size changes
        private void PageForm_SizeChanged(object sender, EventArgs e)
        {
            // make sure robot form's size stays in sync with the tab page
            if (_formBase != null)
            {
                _formBase.Size = (sender as TabPage).Size;
            }
        }

        // called when "Settings" tab's "Robot Year" dropdown's selection changes
        private void comboBoxYear_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (comboBoxYear.SelectedIndex < 0)
            {
                return;
            }

            // the newest year is 1511
            if (comboBoxYear.SelectedIndex == 0)
            {
                textBoxTeam.Text = "1511";
                return;
            }

            // prior years have are YY15
            textBoxTeam.Text = comboBoxYear.SelectedItem.ToString().Substring(2) + "15";
        }

        // called with "Settings" tab's "Set" button is pressed
        private void buttonConfiguration_Click(object sender, EventArgs e)
        {
            if (comboBoxYear.SelectedIndex < 0)
            {
                MessageBox.Show("Select a Robot Year");
                return;
            }
            if (textBoxTeam.TextLength != 4)
            {
                MessageBox.Show("Pick a 4 digit team year");
            }
            try
            {
                Properties.Settings.Default.Year = comboBoxYear.SelectedItem.ToString();
                Properties.Settings.Default.Team = textBoxTeam.Text;
                Properties.Settings.Default.Save();
            }
            catch
            {
                return;
            }

            // launch a new one that will use the new settings
            System.Diagnostics.Process process = new System.Diagnostics.Process();
            string file = System.Reflection.Assembly.GetEntryAssembly().GetName().CodeBase;
            process.StartInfo = new System.Diagnostics.ProcessStartInfo(file);
            process.Start();

            // now we can close ourselves
            Close();
        }

        // handle the debug list being double-clicked
        private void ListViewRoboDebug_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            DashboardNetworkTable.EditItem(ref listViewRoboDebug, e);
        }

        // called when 'hidden' control to show net tbl debug is double-clicked
        private void panelNetTbl_DoubleClick(object sender, EventArgs e)
        {
            groupBoxNetTbl.Visible = !groupBoxNetTbl.Visible;
        }

        // set the background colors of controls to indicate what is going on
        private void SetBackColors()
        {
            Color color = Color.Transparent;

            // gyro calibration wins over EVERYTHING
            if (_gyroCalibrate)
            {
                color = Color.Red;
            }
            else if (_inPitMode)
            {
                // when in pit mode we want the color be something 'strange'
                color = Color.PowderBlue;

                if (_colorYellow && !_colorGreen)
                {
                    color = Color.Plum;                 // warning only
                }
                else if (!_colorYellow && _colorGreen)
                {
                    color = Color.Orange;
                }
                else if (_colorYellow && _colorGreen)
                {
                    color = Color.DarkViolet;          // warning and 'good'
                }
            }
            else
            {
                // normal mode
                if (_colorYellow && !_colorGreen)
                {
                    color = Color.Yellow;               // warning only
                }
                else if (!_colorYellow && _colorGreen)
                {
                    color = Color.Green;                // 'good' only
                }
                else if (_colorYellow && _colorGreen)
                {
                    color = Color.YellowGreen;          // warning and 'good'
                }
            }

            // set all the main tab pages, as well as the aux/driver tabs
            if (_formBase != null)
            {
                _formBase.BackColor = ((color == Color.Transparent) ? SystemColors.Control : color);
            }
            foreach (TabPage tabPage in tabControlDashboard.TabPages)
            {
                tabPage.BackColor = color;
            }
            tabPageAuxIOMap.BackColor = color;
            tabPageDriverIOMap.BackColor = color;
        }

        // called when a entry in the table changes
        private void NetworkTableDelegate(string key, string value)
        {
            // validate
            if ((key == null) || (value == null))
            {
                return;
            }

            // make sure the call happens in the UI thread
            if (InvokeRequired)
            {
                BeginInvoke(new DashboardNetworkTable.DashboardNetworkTableDelegate(NetworkTableDelegate), key, value);
                return;
            }

            // see if it is for the main robot form
            if ((_formBase != null) && _formBase.NetworkTableUpdated(key, value))
            {
                // robot form handled all the processing of this key, so we are done
                return;
            }

            // see which item it is in the network table that we have
            switch (key)
            {
                default:
                    // see if it is a chart key
                    if (ChartValue(key, value))
                    {
                        break;
                    }

                    // nothing we care about
                    break;

                case "thunderdashboard_inpitmode":
                    _inPitMode = (value == "1");
                    SetBackColors();    // make sure background colors stay in sync
                    break;

                case "thunderdashboard_gyro":
                    _gyroCalibrate = (value == "1");
                    SetBackColors();    // make sure background colors stay in sync
                    break;

                case "thunderdashboard_yellow":
                    _colorYellow = (value == "1");
                    SetBackColors();    // make sure background colors stay in sync
                    break;

                case "thunderdashboard_green":
                case "thunderdashboard_haveball": // this is 2016 specific, but needed for backward compatiblity
                    _colorGreen = (value == "1");
                    SetBackColors();    // make sure background colors stay in sync
                    break;
            }

            // update the robot debug table
            if (!DashboardNetworkTable.ShowTableItem(ref listViewRoboDebug, key, value))
            {
                if ((_formBase != null) && (_formBase.StatusTextBox != null))
                {
                    string netTblError = DashboardNetworkTable.LastError();
                    if (netTblError != "")
                    {
                        _formBase.StatusTextBox.Text = netTblError;
                    }
                }
            }
        }


        // member variables
        FormBase _formBase = null; // form for robot year we are running
        private bool _runAsServer = false; // are we running a 'robot' (network table server)

        private bool _inPitMode = false; // robot wants dashboard background to indicate we are in pit mode?
        private bool _gyroCalibrate = false; // robot wants dashboard background to indicate gyro is calibrating?
        private bool _colorYellow = false; // robot wants dashboard background to be yellow?
        private bool _colorGreen = false; // robot wants dashboard background to be green?
    }
}