////////////////////////////////////////////////////////////////////////////////////////////////////
//
//	The order for processing an image:
//		1. ConvertRgbToMono			pulls the green from the camera image and creates monochrome
//										image containing just the green targets
//		2. CalcTargetRows			uses mono image to determine vertically where the targets are
//		3. CalcTargetColumns		uses mono image and CalcTargetRows() results to determine
//										horizontally where each target is
//		4. CalcTargets				uses mono image and CalcTargetColumns() results to 'tighten'
//										the boxes for each target and calc final values
//		OR
//		ProcessImage				will process the entire image all at once (i.e. run STEPs 1 - 4)
//
//	After processing:
//		GetTargetInfo				gets list of targets found
//
//	History:
//		03/04/12 - now skips targets that are too small (i.e. added MIN_HEIGHT and MIN_WIDTH)
//					NOTE: set both MIN_ to zero to get previous behavior
//
////////////////////////////////////////////////////////////////////////////////////////////////////

/// //////////////////////////////////////////////////////////
// for C++, uncomment the lines in this section
//#define CPP
//#ifndef THUNDERVISION
//#define THUNDERVISION

//#include "WPILib.h"
//#include "nivision.h"
//typedef unsigned char byte;
//////////////////////////////////////////////////////////////

#if !CPP
#define PC_TEST_APP
namespace ThunderDashboard
{
    public partial class ThunderVision
    {
        public struct RGBValue
        {
            public byte B;     //The blue value of the color.
            public byte G;     //The green value of the color.
            public byte R;     //The red value of the color.
            public byte alpha; //The alpha value of the color, which represents extra information about a color image, such as gamma correction.
        };

#else
	class ThunderVision
	{
public:
		ThunderVision();
		~ThunderVision();
#endif

        // information we need to track for each target
        public struct TARGET_INFO
        {
            public bool valid;                      // valid target (usually set when image is processed)
            public int row;                        // row within the image (0 is first row)
            public int left;                       // left-most pixel
            public int right;                      // right-most pixel
            public int top;                        // top-most pixel
            public int bottom;                     // bottom-most pixel
            public int avgWidth;                   // average width
            public int avgHeight;                  // average height
            public int maxWidth;                   // max width found
            public int maxHeight;                  // max height found
            public int edgeLeft;                   // 'angle' of left edge (0 = straight; neg = \; pos = /)
            public int edgeTop;                    // 'angle' of top edge (0 = straight; neg = /; pos = \)
        };

#if CPP
        // will process the entire image all at once (i.e. run STEPs 1 through 4)
        //
        // returns the number of targets found
        int ProcessImage(RGBValue *image, int width, int height);

		// convert the given RGB image (from the camera) into a monochrome image, by pulling
		// out the green tape reflections. the mono image is stored within this class
		//
		// the image MUST be larger than 640 x 480
		//
		// returns:
		//    false - image is not valid; DO NOT CALL ANYMORE STEPS
		//    true  - image is valid; ok to call rest of the steps
		bool ConvertRgbToMono(RGBValue *image, int width, int height);	// STEP 1

																		// look at each row to see if it has any pixels in it.
																		// any successive lines (with pixels) will be considered part of the same target
		void CalcTargetRows();											// STEP 2

																		// look at each row (from CalcTargetRows()) and find all each box
		void CalcTargetColumns();										// STEP 3

																		// look at each target from CalcTargetColumns() and 'fine tune' it and get final values
		void CalcTargets();												// STEP 4

																		// gets the targets
																		//
																		// targetInfo		is a pointer to a list of targets (pass in NULL to get the total)
																		// numEntires		indicates how many TARGET_INFO structures that targetInfo points to
																		// returns			the number of targets copied
		int GetTargetInfo(TARGET_INFO *targetInfo, int numEntries);

		// for PC test app only
#if PC_TEST_APP
		void GetMonoImage(byte *image) { memcpy(image, monoImage, sizeof(monoImage)); }
		void GetTargetRowImage(RGBValue *image);		// uses CalcTargetRows() results
		void GetTargetImageWithGap(RGBValue *image);	// uses CalcTargetColumns() results (with gaps)
		void GetTargetImageNoGap(RGBValue *image);		// uses CalcTargetColumns() results (w/o gaps)
		void GetTargetImageNoNoise(RGBValue *image);	// uses CalcTargetColumns() results (final w/o gaps, w/o noise)
		void GetTargetImageTight(RGBValue *image);		// uses CalcTargets() result (tight box)
#endif // PC_TEST_APP

        private:
#endif // CPP

        // defines for pulling the green from the RGB camera image
        const int GREEN_THRESHOLD = 110;    // minimum value Green must be
        const int RED_DELTA = 50;           // Red must be this far away from Green [ignore if (r >= (g-RED_DELTA))]
        const int BLUE_THRESHOLD = 40;      // minimum value Blue must be
        const int BLUE_DELTA = 10;          // Blue cannot be this much brigher than Green [ignore if (b > (g+BLUE_DELTA))]

        const int MIN_PIXELS_ROW_320 = 3;   // minimum pixels-per-row we need (use to ignore diagonal lines and noise)
        const int MIN_PIXELS_ROW_640 = 5;   // minimum pixels-per-row we need (use to ignore diagonal lines and noise)
        const int MIN_HEIGHT_320 = 15;      // minimum height a target must be (in pixels)
        const int MIN_HEIGHT_640 = 20;      // minimum height a target must be (in pixels)
        const int MIN_WIDTH_320 = 15;       // minimum width a target must be (in pixels)
        const int MIN_WIDTH_640 = 20;		// minimum width a target must be (in pixels)

        struct ROW_INFO                         // information we need to track for each row
        {
            public int top;
            public int bottom;
        };

#if CPP
		// initialize member variables
		void Init();

		// create column histogram for the given target row
		void CreateRowHisto(int row);

		// look at row's histogram and find all the targets (just getting 'width' of target)
		void FindTargets(int targetFirstInRow, int gap);

		// use CalcTargetColumns() info to fine tune the top & bottom, and get avgWidth
		void FineTune(TARGET_INFO *targetInfo);
		void CalcHeightAndTopEdge(TARGET_INFO *targetInfo);
#endif // CPP

#if !CPP
        byte[] monoImage = new byte[640 * 480];     // grayscale image
        int[] histogram = new int[640];            // histogram for width and height
#else
        byte monoImage[640 * 480];			// grayscale image
        int histogram[640];					// histogram for width and height
#endif

        const int MAX_ROWS = 8;					// max rows we will track (should only be 3, but need more for processing noise)
#if !CPP
        ROW_INFO[] rowInfo = new ROW_INFO[MAX_ROWS];    // all row info
#else
        ROW_INFO rowInfo[MAX_ROWS];         // all row info
#endif
        int rows;                               // number of rows that have targets (i.e. valid items in rowInfo[])

        const int MAX_TARGETS = 16;             // max targets we will track (should only be 4, but need more for processing noise)
#if !CPP
        public TARGET_INFO[] targetInfo = new TARGET_INFO[MAX_TARGETS];    // all target info
#else
        TARGET_INFO targetInfo[MAX_TARGETS];    // all target info
#endif
        public int targets;                            // number of targets (i.e. valid items in targetInfo[])

        int _imageWidth;
        int _imageHeight;
        int _pixelPerRow;
        int _minWidth;
        int _minHeight;

        // for PC test app only
#if PC_TEST_APP

#if CPP
        void GetTargetImage(RGBValue *image, int maxTargets, TARGET_INFO *targetInfo);
#endif

	const int MAX_TARGETS_TEST = 64;            // list of targets before we 'remove' the gaps
#if CPP
        TARGET_INFO targetInfoTest[MAX_TARGETS_TEST];
#else
        TARGET_INFO[] targetInfoTest = new TARGET_INFO[MAX_TARGETS_TEST];
#endif
        int targetsTest;

	const int MAX_TARGETS_TEST_G = 64;          // list of targets w/o gaps but w/noise
#if CPP
    	TARGET_INFO targetInfoTestG[MAX_TARGETS_TEST_G];
#else
        TARGET_INFO[] targetInfoTestG = new TARGET_INFO[MAX_TARGETS_TEST_G];
#endif
        int targetsTestG;

#endif // PC_TEST_APP
#if !CPP
    }
}
#else
    };
#endif //CPP

//#endif // THUNDERVISION